import logging
import os
import subprocess

from hhd.i18n import _

# TODO: Flip this to 0 on release
HHD_SWAP_CREATE = os.environ.get("HHD_SWAP_CREATE", "0") == "1"
HHD_SWAP_SUBVOL = os.environ.get("HHD_SWAP_SUBVOL", "/var/swap")
HHD_SWAP_FILE = os.environ.get("HHD_SWAP_FILE", "/var/swap/hhdswap")

SAFETY_BUFFER = 1.3
ZRAM_MULTIPLIER = 1.5

logger = logging.getLogger(__name__)


def get_windows_bootnum() -> int | None:
    try:
        s = subprocess.check_output("efibootmgr").decode("utf-8")

        for line in s.split("\n"):
            if "Windows Boot Manager" in line:
                return int(line[: line.index(" ")].replace("*", "").replace("Boot", ""))

        return None
    except Exception as e:
        return None


def boot_windows():
    bootnum = get_windows_bootnum()

    if bootnum is None:
        logger.error("Could not find Windows Boot Manager in efibootmgr output")
        return

    try:
        subprocess.run(["efibootmgr", "-n", str(bootnum)])
        logger.info(f"Booting Windows with bootnum {bootnum}")
        subprocess.run(["systemctl", "reboot"])
    except Exception as e:
        logger.error(f"Failed to boot Windows: {e}")


def is_btrfs(fn):
    print()
    return (
        subprocess.run(
            ["stat", "-f", "-c", "%T", fn],
            check=True,
            stdout=subprocess.PIPE,
            text=True,
        ).stdout.strip()
        == "btrfs"
    )


def create_subvol():
    # First, create subvol dir to make checks
    os.makedirs(HHD_SWAP_SUBVOL, exist_ok=True)

    # Check filesystem is btrfs
    if not is_btrfs(HHD_SWAP_SUBVOL):
        logger.info("Swap filesystem is not btrfs. Skipping subvolume creation.")
        return

    # Check if subvolume already exists
    if (
        subprocess.run(
            ["btrfs", "subvolume", "show", HHD_SWAP_SUBVOL],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ).returncode
        == 0
    ):
        logger.info(
            f"Swap subvolume {HHD_SWAP_SUBVOL} already exists. Skipping creation."
        )
        return

    # Fixup selinux for swap
    subprocess.run(
        ["semanage", "fcontext", "-a", "-t", "var_t", HHD_SWAP_SUBVOL],
    )
    subprocess.run(["restorecon", HHD_SWAP_SUBVOL])

    logger.info(f"Creating swap subvolume {HHD_SWAP_SUBVOL}")
    os.system(f"btrfs subvolume create {HHD_SWAP_SUBVOL}")


def get_meminfo():
    with open("/proc/meminfo") as f:
        lines = f.readlines()

    meminfo = {}
    for line in lines:
        key, value = line.split(":")
        meminfo[key.strip()] = int(value.strip().split(" ")[0])

    return meminfo


def create_temporary_swap():
    if not HHD_SWAP_CREATE:
        return

    swapdata = subprocess.run(
        ["swapon", "--show", "--raw"], capture_output=True, text=True
    ).stdout.split("\n")[1:]

    # Check if there is a swap on disk
    has_swap = any(["/dev/zram" not in line and line for line in swapdata])
    assert not has_swap, (
        "Found swap partition. We cannot create temporary swap. Bail second attempt. Swap output:\n"
        + "\n".join(swapdata)
    )

    if HHD_SWAP_SUBVOL:
        create_subvol()

    # Check if there is a ZRAM swap partition
    has_zram = any(["/dev/zram" in line for line in swapdata])
    meminfo = get_meminfo()

    required_kb = meminfo["MemTotal"] - meminfo["MemFree"]
    required_kb *= SAFETY_BUFFER
    # ZRAM can compress a lot, add a safety buffer
    if has_zram:
        required_kb *= ZRAM_MULTIPLIER

    if os.path.exists(HHD_SWAP_FILE):
        os.remove(HHD_SWAP_FILE)

    if is_btrfs(os.path.dirname(HHD_SWAP_FILE)):
        logger.info(f"Creating BTRFS swapfile {HHD_SWAP_FILE}")
        subprocess.run(
            [
                "btrfs",
                "filesystem",
                "mkswapfile",
                HHD_SWAP_FILE,
                "--size",
                f"{int(required_kb)}k",
            ],
            check=True,
        )
    else:
        logger.info(f"Creating swapfile {HHD_SWAP_FILE} (w fallocate/mkswap)")
        subprocess.run(
            ["fallocate", "-l", f"{int(required_kb)}K", HHD_SWAP_FILE], check=True
        )
        subprocess.run(["chmod", "600", HHD_SWAP_FILE], check=True)
        subprocess.run(["mkswap", HHD_SWAP_FILE], check=True)

    # Fixup selinux for swap
    subprocess.run(
        [
            "semanage",
            "fcontext",
            "-a",
            "-t",
            "swapfile_t",
            HHD_SWAP_FILE,
        ],
    )
    subprocess.run(["restorecon", HHD_SWAP_FILE])

    # Enable swap
    subprocess.run(["swapon", HHD_SWAP_FILE], check=True)

    # Reset resume device so systemd does not get confused
    with open("/sys/power/resume", "w") as f:
        f.write("0:0")
    with open("/sys/power/resume_offset", "w") as f:
        f.write("0")

    # Disable zram to avoid confusing the kernel
    # Systemd will re-enable it after hibernation
    for zram in swapdata:
        if "/dev/zram" not in zram:
            continue

        zram = zram.strip().split(" ")[0]
        logger.info(f"Disabling ZRAM swap {zram}")
        subprocess.run(["swapoff", zram], check=True)


def delete_temporary_swap():
    if not HHD_SWAP_CREATE:
        return

    if not os.path.exists(HHD_SWAP_FILE):
        return

    logger.info(f"Deleting swapfile {HHD_SWAP_FILE}")
    subprocess.run(["swapoff", HHD_SWAP_FILE], check=True, stdout=subprocess.DEVNULL)
    os.remove(HHD_SWAP_FILE)


def emergency_shutdown():
    logger.error("HIBERNATION FAILED. INITIATING EMERGENCY SHUTDOWN.")
    os.system("systemctl poweroff")


def supports_sleep():
    # https://gitlab.freedesktop.org/drm/amd/-/blob/master/scripts/amd_s2idle.py
    try:
        fn = os.path.join("/", "sys", "power", "mem_sleep")
        if not os.path.exists(fn):
            logger.error(
                "Kernel compiled without sleep support. Sleep button will force hibernate."
            )
            return False

        with open(fn) as f:
            sleep = f.read().strip()

        if "deep" in sleep:
            logger.info("S3 sleep supported, sleep button will work.")
            return True

        import struct

        target = os.path.join("/", "sys", "firmware", "acpi", "tables", "FACP")
        with open(target, "rb") as r:
            r.seek(0x70)
            BIT = lambda x: 1 << x
            found = struct.unpack("<I", r.read(4))[0] & BIT(21)
            s2idle_supported = bool(found)

        if s2idle_supported:
            logger.info("S2idle sleep supported, sleep button will work.")
        else:
            logger.error(
                "S2idle sleep not supported. Sleep button will force hibernate."
            )

        return s2idle_supported

    except Exception as e:
        logging.error(f"Failed to read FADT: {e}. Assuming sleep is supported")
    return True


def emergency_hibernate(shutdown: bool = False):
    # Try to hibernate with built in swap
    logger.warning("Commencing emergency hibernation")

    logger.info("Dropping caches")
    with open("/proc/sys/vm/drop_caches", "w") as f:
        f.write("3")

    logger.info("Hibernating")
    # Here loginctl does a soft swap check and errors
    # out if there is no swap
    ret = os.system("systemctl hibernate")
    if not ret:
        return ""

    status = _("Failed to hibernate (missing swap file).")

    if HHD_SWAP_CREATE:
        # Create temporary swap and hibernate
        logger.warning(
            "Hibernation failed (presumably no swap), creating temporary swap and hibernating"
        )
        try:
            create_temporary_swap()
            created = True
        except Exception as e:
            logger.error(f"Failed to create temporary swap:\n{e}")
            created = False
            status = _("Failed to create temporary swap.")

        if created:
            ret = os.system("systemctl hibernate")

            # If this fails we have to emergency shutdown
            # FIXME: That function returns before hibernation is complete
            if not ret:
                return ""

            status = _("Failed to hibernate to temporary swap.")

    if shutdown:
        emergency_shutdown()
    return status


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    HHD_SWAP_CREATE = True
    emergency_hibernate(shutdown=False)



from hhd.controller import Axis, Button, Configuration
from hhd.controller.physical.evdev import B, to_map
from hhd.plugins import gen_gyro_state

DEFAULT_MAPPINGS: dict[str, tuple[Axis, str | None, float, float | None]] = {
    "accel_x": ("accel_z", "accel", 1, None),
    "accel_y": ("accel_x", "accel", -1, None),
    "accel_z": ("accel_y", "accel", -1, None),
    "anglvel_x": ("gyro_z", "anglvel", 1, None),
    "anglvel_y": ("gyro_x", "anglvel", -1, None),
    "anglvel_z": ("gyro_y", "anglvel", -1, None),
    "timestamp": ("imu_ts", None, 1, None),
}

X1_MAPPING = gen_gyro_state("x", True, "z", False, "y", False)
X1_MINI_MAPPING = gen_gyro_state("z", True, "x", False, "y", True)

BTN_MAPPINGS: dict[int, Button] = {
    # Volume buttons come from the same keyboard
    B("KEY_VOLUMEUP"): "key_volumeup",
    B("KEY_VOLUMEDOWN"): "key_volumedown",
    # Turbo Button [29, 56, 125] KEY_LEFTCTRL + KEY_LEFTALT + KEY_LEFTMETA
    B("KEY_LEFTALT"): "share",
    # Short press orange [32, 125] KEY_D + KEY_LEFTMETA
    B("KEY_D"): "mode",
    # KB Button [24, 97, 125]  KEY_O + KEY_RIGHTCTRL + KEY_LEFTMETA
    B("KEY_O"): "keyboard",
}

BTN_MAPPINGS_NONTURBO: dict[int, Button] = {
    # Volume buttons come from the same keyboard
    B("KEY_VOLUMEUP"): "key_volumeup",
    B("KEY_VOLUMEDOWN"): "key_volumedown",
    # Short press orange [32, 125] KEY_D + KEY_LEFTMETA
    B("KEY_D"): "mode",
    # KB Button [24, 97, 125]  KEY_O + KEY_RIGHTCTRL + KEY_LEFTMETA
    # If we do not have turbo takeover, let turbo do its turbo thing, and
    # failover to having the keyboard button open the overlay
    B("KEY_O"): "share",
}

ONEX_DEFAULT_CONF = {
    "hrtimer": True,
}

OXP_F1_CONF = {
    "name": "ONEXPLAYER ONEXFLY",
    **ONEX_DEFAULT_CONF,
    "protocol": "mixed",
}
OXP_2_CONF = {
    "name": "ONEXPLAYER 2",
    **ONEX_DEFAULT_CONF,
    "protocol": "mixed",
    "rgb": False,
    "buttons": "none",
    "protocol": "none",  # explicitly disable
}
AOKZOE_CONF = {
    "name": "AOKZOE A1",
    "hrtimer": True,
    "protocol": "none",
    "rgb": False,
}

CONFS = {
    # Aokzoe
    "AOKZOE A1 AR07": AOKZOE_CONF,
    "AOKZOE A1 Pro": AOKZOE_CONF,
    # Onexplayer
    "ONE XPLAYER": {"name": "ONE XPLAYER", **ONEX_DEFAULT_CONF},
    "ONEXPLAYER Mini Pro": {
        "name": "ONEXPLAYER Mini Pro",
        **ONEX_DEFAULT_CONF,
        "protocol": "hid_v2",
    },
    "ONEXPLAYER mini A07": {"name": "ONEXPLAYER mini", **ONEX_DEFAULT_CONF},
    # OneXFly
    "ONEXPLAYER F1": OXP_F1_CONF,
    "ONEXPLAYER F1 EVA-01": OXP_F1_CONF,
    "ONEXPLAYER F1L": OXP_F1_CONF,
    "ONEXPLAYER F1 OLED": OXP_F1_CONF,
    "ONEXPLAYER F1Pro": OXP_F1_CONF,
    "ONEXPLAYER F1 EVA-02": OXP_F1_CONF,  # F1Pro variant
    # OXP 2
    "ONEXPLAYER 2": OXP_2_CONF,
    "ONEXPLAYER 2 ARP23": OXP_2_CONF,
    "ONEXPLAYER 2 GA18": OXP_2_CONF,
    # Pro is a bit different
    "ONEXPLAYER 2 PRO ARP23": OXP_2_CONF,
    "ONEXPLAYER 2 PRO ARP23 EVA-01": OXP_2_CONF,
    # X1 Line
    "ONEXPLAYER X1 mini": {
        **ONEX_DEFAULT_CONF,
        "name": "ONEXPLAYER X1 mini",
        "x1": True,
        "mapping": X1_MINI_MAPPING,
        "protocol": "hid_v1",
    },
    "ONEXPLAYER X1 A": {
        **ONEX_DEFAULT_CONF,
        "name": "ONEXPLAYER X1 (AMD)",
        "x1": True,
        "rgb_secondary": True,
        "mapping": X1_MAPPING,
        "protocol": "serial",
    },
    "ONEXPLAYER X1Pro": {
        **ONEX_DEFAULT_CONF,
        "name": "ONEXPLAYER X1 (AMD)",
        "x1": True,
        "rgb_secondary": True,
        "mapping": X1_MAPPING,
        "protocol": "serial",
    },
    "ONEXPLAYER X1 i": {
        **ONEX_DEFAULT_CONF,
        "name": "ONEXPLAYER X1 (Intel)",
        "x1": True,
        "rgb_secondary": True,
        "mapping": X1_MAPPING,
        "protocol": "serial",
        "turbo": False,  # disable turbo takeover so that it can be used for TDP
    },
}


def get_default_config(product_name: str, manufacturer: str):
    out = {
        "name": product_name,
        "manufacturer": manufacturer,
        "hrtimer": True,
        "untested": True,
        "x1": "X1" in product_name,
    }

    if "X1" in product_name and "mini" not in product_name.lower():
        out["rgb_secondary"] = True

    if "aokzoe" in manufacturer.lower():
        out["protocol"] = "none"
        out["rgb"] = False

    return out



import copy
import os
import subprocess
from gettext import GNUTranslations, find
from typing import Mapping, Sequence, cast

from hhd.plugins import Config, Context, HHDLocale, HHDSettings

_translations = {}


def get_user_lang(ctx: Context):
    if not ctx:
        return None
    try:
        out = subprocess.check_output(
            ["sh", "-l", "-c", "locale"],
            env={},
            user=ctx.euid,
            group=ctx.egid,
        )
        for ln in out.decode().split("\n"):
            if "LANG" in ln:
                return ln.strip().split("=")[-1]
        # Alternative:
        # return subprocess.check_output(
        #     ["sh", "-l", "-c", "echo $LANG"],
        #     env={},
        #     user=ctx.euid,
        #     group=ctx.egid,
        # ).decode()
    except Exception:
        return None


def translate_ver(conf: Config, lang: str | None = None, user_lang: str | None = None):
    v = conf.get("version", "")

    if not lang:
        lang = conf.get("hhd.settings.language", "")
    if lang == "system" and user_lang:
        lang = user_lang
    return v + "-" + lang


def get_mo_files(
    conf: Config,
    locales: Sequence[HHDLocale],
    lang: str | None = None,
    user_lang: str | None = None,
):
    if not lang:
        lang = conf.get("hhd.settings.language", "")
    if lang == "system" and user_lang:
        lang = user_lang
    if lang and lang != "system":
        languages = [lang]
    else:
        languages = None

    fns = []
    for locale in locales:
        fns.extend(find(locale["domain"], locale["dir"], languages, all=True))
    return fns


def translation(
    conf: Config,
    locales: Sequence[HHDLocale],
    lang: str | None = None,
    user_lang: str | None = None,
):
    mofiles = get_mo_files(conf, locales, lang, user_lang)
    result = None
    for mofile in mofiles:
        key = (GNUTranslations, os.path.abspath(mofile))
        t = _translations.get(key)
        if t is None:
            with open(mofile, "rb") as fp:
                t = _translations.setdefault(key, GNUTranslations(fp))

        t = copy.copy(t)
        if result is None:
            result = t
        else:
            result.add_fallback(t)
    return result


def trn_dict(d: Mapping, trn: GNUTranslations):
    out = dict(d)
    for k, v in d.items():
        if isinstance(v, dict):
            out[k] = trn_dict(v, trn)
        elif isinstance(v, str) and v:
            out[k] = trn.gettext(v)
        elif isinstance(v, list):
            out[k] = [trn.gettext(l) if l and isinstance(l, str) else l for l in v]
        else:
            out[k] = v
    return out


def translate(
    d: Mapping,
    conf: Config,
    locales: Sequence[HHDLocale],
    lang: str | None = None,
    user_lang: str | None = None,
):
    trn = translation(conf, locales, lang, user_lang)
    base = d
    if trn:
        base = trn_dict(base, trn)
    return base



import logging
from typing import Any, Literal, Mapping, Sequence

import os
from ..controller.base import Consumer, Producer, RgbMode, RgbSettings, RgbZones
from ..controller.virtual.dualsense import Dualsense, TouchpadCorrectionType
from ..controller.virtual.uinput import (
    CONTROLLER_THEMES,
    GAMEPAD_BUTTON_MAP,
    HHD_PID_TOUCHPAD,
    HORIPAD_STEAM_BUTTON_MAP,
    MOTION_AXIS_MAP,
    MOTION_AXIS_MAP_FLIP_Z,
    MOTION_CAPABILITIES,
    MOTION_INPUT_PROPS,
    TOUCHPAD_AXIS_MAP,
    TOUCHPAD_BUTTON_MAP,
    TOUCHPAD_CAPABILITIES,
    XBOX_ELITE_BUTTON_MAP,
    UInputDevice,
)
from .plugin import is_steam_gamepad_running, open_steam_kbd
from .utils import load_relative_yaml

logger = logging.getLogger(__name__)

HORI_ENABLED = os.environ.get("HHD_HORI_STEAM", "0") == "1"


def get_outputs(
    conf,
    touch_conf,
    motion: bool = False,
    *,
    controller_id: int = 0,
    emit=None,
    dual_motion: bool = False,
    rgb_modes: Mapping[RgbMode, Sequence[RgbSettings]] | None = None,
    rgb_zones: RgbZones = "mono",
    controller_disabled: bool = False,
) -> tuple[Sequence[Producer], Sequence[Consumer], Mapping[str, Any]]:
    producers = []
    consumers = []
    nintendo_qam = False

    controller = conf["mode"].to(str)
    desktop_disable = False
    if touch_conf is not None:
        touchpad = touch_conf["mode"].to(str)
        correction = touch_conf["controller.correction"].to(TouchpadCorrectionType)
        if touchpad in ("emulation", "controller"):
            desktop_disable = touch_conf[touchpad]["desktop_disable"].to(bool)
    else:
        touchpad = "controller"
        correction = "stretch"

    # Run steam check for touchpad
    steam_check = (
        is_steam_gamepad_running(emit.ctx) if emit and desktop_disable else None
    )
    match steam_check:
        case True:
            logger.info("Gamepadui active. Activating touchpad emulation.")
        case False:
            logger.info("Gamepadui closed. Disabling touchpad emulation.")

    has_qam = False
    uses_touch = False
    uses_leds = False
    noob_mode = False
    flip_z = False

    if controller_disabled:
        controller = "hidden"

    match controller:
        case "hidden":
            # NOOP
            UInputDevice.close_cached()
            Dualsense.close_cached()
            motion = False
            noob_mode = conf.get("hidden.noob_mode", False)
        case "dualsense":
            UInputDevice.close_cached()
            flip_z = conf["dualsense.flip_z"].to(bool)
            uses_touch = touchpad == "controller" and steam_check is not False
            uses_leds = conf.get("dualsense.led_support", False)
            paddles_as = conf.get("dualsense.paddles_as", "noob")
            noob_mode = paddles_as in ("noob", "both")
            edge_mode = False

            if paddles_as == "both":
                paddles_to_clicks = "bottom"
            elif paddles_as == "touchpad":
                paddles_to_clicks = "top"
            elif paddles_as == "steam_input":
                edge_mode = True
                paddles_to_clicks = "disabled"
            else:
                paddles_to_clicks = "disabled"

            d = Dualsense(
                touchpad_method=correction,
                edge_mode=edge_mode,
                use_bluetooth=conf["dualsense.bluetooth_mode"].to(bool),
                enable_touchpad=uses_touch,
                enable_rgb=uses_leds,
                fake_timestamps=not motion,
                sync_gyro=conf["dualsense.sync_gyro"].to(bool) and motion,
                paddles_to_clicks=paddles_to_clicks,
                flip_z=flip_z,
                controller_id=controller_id | (0xF0 if edge_mode else 0),
                cache=True,
            )
            producers.append(d)
            consumers.append(d)
        case "uinput" | "xbox_elite" | "joycon_pair" | "hori_steam":
            Dualsense.close_cached()
            version = 1
            sync_gyro = False
            paddles_as = conf.get("uinput.paddles_as", "noob")
            if controller == "joycon_pair":
                theme = "joycon_pair"
                nintendo_qam = conf["joycon_pair.nintendo_qam"].to(bool)
                button_map = GAMEPAD_BUTTON_MAP
                bus = 0x06
                version = 0
            elif controller == "hori_steam":
                theme = "hori_steam"
                noob_mode = conf.get("hori_steam.noob_mode", False)
                flip_z = conf["hori_steam.flip_z"].to(bool)
                button_map = HORIPAD_STEAM_BUTTON_MAP
                bus = 0x06
                version = 0
                sync_gyro = conf.get("hori_steam.sync_gyro", True)
                has_qam = True
            elif controller == "xbox_elite" or (
                controller == "uinput" and paddles_as == "steam_input"
            ):
                theme = "xbox_one_elite"
                button_map = XBOX_ELITE_BUTTON_MAP
                bus = 0x03
            else:
                noob_mode = paddles_as == "noob"
                # theme = conf.get("uinput.theme", "hhd")
                theme = "hhd"
                nintendo_qam = conf["uinput.nintendo_qam"].to(bool)
                # flip_z = conf["uinput.flip_z"].to(bool)
                flip_z = False
                button_map = GAMEPAD_BUTTON_MAP
                bus = 0x03 if theme == "hhd" else 0x06
            vid, pid, name = CONTROLLER_THEMES[theme]
            addr = "phys-hhd-main"
            if controller_id:
                addr = f"phys-hhd-{controller_id:02d}"
            d = UInputDevice(
                name=name,
                vid=vid,
                pid=pid,
                phys=addr,
                uniq=addr,
                btn_map=button_map,
                bus=bus,
                version=version,
                cache=True,
                sync_gyro=sync_gyro and motion,
            )
            producers.append(d)
            consumers.append(d)
            # Deactivate motion if using an xbox theme
            motion = (theme != "hhd" and "xbox" not in theme) and motion
            if motion:
                d = UInputDevice(
                    name=f"{name} Motion Sensors",
                    vid=vid,
                    pid=pid,
                    phys=addr,
                    uniq=addr,
                    bus=bus,
                    version=version,
                    capabilities=MOTION_CAPABILITIES,
                    btn_map={},
                    axis_map=(MOTION_AXIS_MAP_FLIP_Z if flip_z else MOTION_AXIS_MAP),
                    output_imu_timestamps=True,
                    input_props=MOTION_INPUT_PROPS,
                    ignore_cmds=True,
                    cache=True,
                    motions_device=True,
                )
                producers.append(d)
                consumers.append(d)
        case _:
            raise RuntimeError(f"Invalid controller type: '{controller}'.")

    dual_motion = motion and dual_motion
    if dual_motion:
        d = Dualsense(
            edge_mode=False,
            use_bluetooth=True,
            enable_touchpad=False,
            enable_rgb=False,
            fake_timestamps=False,
            sync_gyro=True,
            paddles_to_clicks="disabled",
            flip_z=flip_z,
            controller_id=5,
            cache=True,
            left_motion=True,
        )
        producers.append(d)
        consumers.append(d)

    if touchpad == "emulation" and steam_check is not False and not controller_disabled:
        d = UInputDevice(
            name="Handheld Daemon Touchpad",
            phys="phys-hhd-main",
            capabilities=TOUCHPAD_CAPABILITIES,
            pid=HHD_PID_TOUCHPAD,
            btn_map=TOUCHPAD_BUTTON_MAP,
            axis_map=TOUCHPAD_AXIS_MAP,
            output_timestamps=True,
            ignore_cmds=True,
        )
        producers.append(d)
        consumers.append(d)
        uses_touch = True

    return (
        producers,
        consumers,
        {
            "uses_touch": uses_touch,
            "rgb_used": uses_leds,
            "rgb_modes": rgb_modes,
            "rgb_zones": rgb_zones,
            "is_dual": False,
            "steam_check": steam_check,
            "steam_check_fn": lambda: emit and is_steam_gamepad_running(emit.ctx),
            "steam_kbd": lambda open: open_steam_kbd(emit, open),
            "nintendo_qam": nintendo_qam,
            "uses_motion": motion,
            "uses_dual_motion": dual_motion,
            "noob_mode": noob_mode,
            "has_qam": has_qam,
            "supports_qam": not controller_disabled and controller != "hidden",
        },
    )


def get_outputs_config(
    can_disable: bool = False,
    has_leds: bool = True,
    start_disabled: bool = False,
    default_device: str | None = None,
    extra_buttons: Literal["none", "dual", "quad"] = "dual",
):
    s = load_relative_yaml("outputs.yml")
    try:
        if not can_disable:
            del s["modes"]["disabled"]
        if not has_leds:
            del s["modes"]["dualsense"]["children"]["led_support"]

        if extra_buttons == "none":
            del s["modes"]["dualsense"]["children"]["paddles_as"]
            del s["modes"]["uinput"]["children"]["paddles_as"]
            del s["modes"]["hidden"]["children"]["noob_mode"]
            del s["modes"]["xbox_elite"]
        elif extra_buttons == "dual":
            del s["modes"]["dualsense"]["children"]["paddles_as"]["options"]["both"]

        if HORI_ENABLED:
            # Replace xbox elite with hori
            try:
                del s["modes"]["xbox_elite"]
            except Exception:
                pass
        else:
            del s["modes"]["hori_steam"]

        # Set xbox as default for now
        s["default"] = "uinput"

        # if default_device:
        #     s["default"] = default_device
        if start_disabled:
            s["default"] = "disabled"
    except Exception as e:
        logger.exception(f"Error fixing outputs:\n{e}")
    return s


def get_limits_config(defaults: dict[str, int] = {}):
    s = load_relative_yaml("limits.yml")
    lims = s["modes"]["manual"]["children"]

    lims["ls_min"]["default"] = defaults.get("s_min", 0)
    lims["ls_max"]["default"] = defaults.get("s_max", 95)
    lims["rs_min"]["default"] = defaults.get("s_min", 0)
    lims["rs_max"]["default"] = defaults.get("s_max", 95)

    lims["rt_min"]["default"] = defaults.get("t_min", 0)
    lims["rt_max"]["default"] = defaults.get("t_max", 95)
    lims["lt_min"]["default"] = defaults.get("t_min", 0)
    lims["lt_max"]["default"] = defaults.get("t_max", 95)

    return s


def get_limits(conf, defaults={}):
    if conf["mode"].to(str) != "manual":
        return defaults

    kconf = conf["manual"]
    for set in ("ls", "rs", "lt", "rt"):
        if kconf[f"{set}_min"].to(int) > kconf[f"{set}_max"].to(int):
            kconf[f"{set}_min"] = kconf[f"{set}_max"].to(int)
    return kconf.to(dict)


def fix_limits(conf, prefix: str, defaults: dict[str, int] = {}):
    if conf[f"{prefix}.mode"].to(str) != "manual":
        return {}

    # Make sure min < max
    for set in ("ls", "rs", "lt", "rt"):
        if conf[f"{prefix}.manual.{set}_min"].to(int) > conf[
            f"{prefix}.manual.{set}_max"
        ].to(int):
            conf[f"{prefix}.manual.{set}_min"] = conf[f"{prefix}.manual.{set}_max"].to(
                int
            )

    # Allow reseting limits
    if conf[f"{prefix}.manual.reset"].to(bool):
        for set in ("l", "r"):
            for comp in ("t", "s"):
                conf[f"{prefix}.manual.{set}{comp}_min"] = defaults.get(
                    f"{comp}_min", 0
                )
                conf[f"{prefix}.manual.{set}{comp}_max"] = defaults.get(
                    f"{comp}_max", 95
                )
        conf[f"{prefix}.manual.reset"] = False



from hhd.plugins import HHDLocale, get_relative_fn


def _(arg: str):
    return arg


def locales() -> list[HHDLocale]:
    return [
        {"dir": get_relative_fn("./"), "domain": "hhd", "priority": 10},
        {"dir": get_relative_fn("./"), "domain": "adjustor", "priority": 20},
    ]



import logging
import os
import time

WAKEUP_DIR = "/sys/class/wakeup"
SMBIOS_FN = "/sys/firmware/dmi/entries/1-0/raw"

def main_loop():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)s] %(message)s",
    )

    logger = logging.getLogger(__name__)

    event_counts = {}
    wakeup_counts = {}
    old_reason = None

    while True:
        for entry in os.scandir(WAKEUP_DIR):
            if not entry.is_dir():
                continue

            with open(os.path.join(entry.path, "event_count"), "r") as f:
                event_count = f.read().strip()

            with open(os.path.join(entry.path, "wakeup_count"), "r") as f:
                wakeup_count = f.read().strip()

            diff = False
            if entry.path not in wakeup_counts:
                wakeup_counts[entry.path] = wakeup_count
                # diff = True

            if wakeup_counts[entry.path] != wakeup_count:
                diff = True
                wakeup_counts[entry.path] = wakeup_count
            
            if entry.path not in event_counts:
                event_counts[entry.path] = event_count
                # diff = True
            
            if event_counts[entry.path] != event_count:
                diff = True
                event_counts[entry.path] = event_count
            
            if not diff:
                continue

            with open(os.path.join(entry.path, "name"), "r") as f:
                name = f.read().strip()

            logger.info(f"{name:>20s}: {wakeup_count:>3s} wakeups, {event_count:>3s} events")
            wakeup_counts[entry.path] = wakeup_count

        with open(SMBIOS_FN, "rb") as f:
            smbios = f.read()

        wakeup_reason = smbios[24]
        # 00h Reserved
        # 01h Other
        # 02h Unknown
        # 03h APM Timer
        # 04h Modem Ring
        # 05h LAN Remote
        # 06h Power Switch
        # 07h PCI PME#
        # 08h AC Power Restored
        match wakeup_reason:
            case 0x00:
                reason = "Reserved"
            case 0x01:
                reason = "Other"
            case 0x02:
                reason = "Unknown"
            case 0x03:
                reason = "APM Timer"
            case 0x04:
                reason = "Modem Ring"
            case 0x05:
                reason = "LAN Remote"
            case 0x06:
                reason = "Power Switch"
            case 0x07:
                reason = "PCI PME#"
            case 0x08:
                reason = "AC Power Restored"
            case _:
                reason = f"Unknown ({wakeup_reason:02X})"

        if old_reason != reason:
            logger.info(f"Wakeup reason: {reason}")
            old_reason = reason

        time.sleep(0.2)


def main():
    try:
        main_loop()
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()



import getpass
import logging
import os
import subprocess
from typing import Any, Literal, Mapping, NamedTuple, Protocol, Sequence, TypedDict

from hhd.controller import Axis, Button, Configuration, ControllerEmitter, SpecialEvent

from .conf import Config
from .settings import HHDSettings

logger = logging.getLogger(__name__)

STEAM_PID = "~/.steam/steam.pid"
STEAM_EXE = "~/.steam/root/ubuntu12_32/steam"


class Context(NamedTuple):
    euid: int = 0
    egid: int = 0
    uid: int = 0
    gid: int = 0
    name: str = "root"
    # scratch: str = ""


class SettingsEvent(TypedDict):
    type: Literal["settings"]


class PowerEvent(TypedDict):
    type: Literal["acpi"]
    event: Literal["ac", "dc", "tdp", "battery"]


class TdpEvent(TypedDict):
    type: Literal["tdp"]
    tdp: int | None


class ProfileEvent(TypedDict):
    type: Literal["profile"]
    name: str
    config: Config | None


class ApplyEvent(TypedDict):
    type: Literal["apply"]
    name: str


class ConfigEvent(TypedDict):
    type: Literal["state"]
    config: Config


class EnergyEvent(TypedDict):
    type: Literal["ppd", "energy"]
    status: Literal["power", "balanced", "performance"]


class InputEvent(TypedDict):
    type: Literal["input"]
    controller_id: int

    btn_state: Mapping[Button, bool]
    axis_state: Mapping[Axis, bool]
    conf_state: Mapping[Configuration, Any]


Event = (
    ConfigEvent
    | InputEvent
    | ProfileEvent
    | ApplyEvent
    | SettingsEvent
    | SpecialEvent
    | PowerEvent
    | TdpEvent
    | EnergyEvent
)


class Emitter(ControllerEmitter):
    def __init__(self, ctx=None, info=None) -> None:
        if info is None:
            info = Config()
        self.info = info
        self.data = {}
        self.images = {}
        super().__init__(ctx)

    def __call__(self, event: Event | Sequence[Event]) -> None:
        pass

    def set_gamedata(
        self, data: dict[str, dict[str, str]], images: dict[str, dict[str, str]]
    ) -> None:
        with self.intercept_lock:
            self.data = data
            self.images = images

    def get_gamedata(self, game: str | None) -> dict[str, str] | None:
        if not game:
            return None
        with self.intercept_lock:
            return self.data.get(game, None)

    def get_image(self, game: str, icon: str) -> str | None:
        with self.intercept_lock:
            return self.images.get(game, {}).get(icon, None)


class HHDPlugin:
    name: str
    priority: int
    log: str

    def open(
        self,
        emit: Emitter,
        context: Context,
    ):
        pass

    def settings(self) -> HHDSettings:
        return {}

    def validate(self, tags: Sequence[str], config: Any, value: Any):
        return False

    def prepare(self, conf: Config):
        pass

    def update(self, conf: Config):
        pass

    def notify(self, events: Sequence[Event]):
        pass

    def close(self):
        pass


class HHDAutodetect(Protocol):
    def __call__(self, existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
        raise NotImplementedError()


class HHDLocale(TypedDict):
    dir: str
    domain: str
    priority: int


class HHDLocaleRegister(Protocol):
    def __call__(self) -> Sequence[HHDLocale]:
        raise NotImplementedError()


def get_context(user: str | None) -> Context | None:
    try:
        uid = os.getuid()
        gid = os.getgid()

        if not user:
            if not uid:
                print(f"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
                print(
                    "Running as root without a specified user (`--user`). Configs will be placed at `/root/.config`."
                )
                print(f"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!")
            return Context(uid, gid, uid, gid, getpass.getuser())

        user = user.replace("\\x2", "-")

        euid = int(
            subprocess.run(
                ["id", "-u", user], capture_output=True, check=True
            ).stdout.decode()
        )
        egid = int(
            subprocess.run(
                ["id", "-g", user], capture_output=True, check=True
            ).stdout.decode()
        )

        if (uid or gid) and (uid != euid or gid != egid):
            print(
                f"The user specified with --user is not the user this process was started with."
            )
            return None

        return Context(euid, egid, uid, gid, user)
    except subprocess.CalledProcessError as e:
        print(f"Getting the user uid/gid returned an error:\n{e.stderr.decode()}")
        return None
    except Exception as e:
        print(f"Failed getting permissions with error:\n{e}")
        return None


def switch_priviledge(p: Context, escalate=False):
    uid = os.geteuid()
    gid = os.getegid()

    if escalate:
        os.seteuid(p.uid)
        os.setegid(p.gid)
    else:
        os.setegid(p.egid)
        os.seteuid(p.euid)

    return uid, gid


def restore_priviledge(old: tuple[int, int]):
    uid, gid = old
    # Try writing group first in case of root
    # and fail silently
    try:
        os.setegid(gid)
    except Exception:
        pass
    os.seteuid(uid)
    os.setegid(gid)
    pass


def expanduser(path: str, user: int | str | Context | None = None):
    """Expand ~ and ~user constructions.  If user or $HOME is unknown,
    do nothing.

    Modified from the python implementation to support using the target userid/user."""

    path = os.fspath(path)

    if not path.startswith("~"):
        return path

    i = path.find("/", 1)
    if i < 0:
        i = len(path)
    if i == 1:
        if "HOME" in os.environ and not user:
            # Fallback to environ only if user not set
            userhome = os.environ["HOME"]
        else:
            try:
                import pwd
            except ImportError:
                # pwd module unavailable, return path unchanged
                return path
            try:
                if not user:
                    userhome = pwd.getpwuid(os.getuid()).pw_dir
                elif isinstance(user, int):
                    userhome = pwd.getpwuid(user).pw_dir
                elif isinstance(user, Context):
                    userhome = pwd.getpwuid(user.euid).pw_dir
                else:
                    userhome = pwd.getpwnam(user).pw_dir
            except KeyError:
                # bpo-10496: if the current user identifier doesn't exist in the
                # password database, return the path unchanged
                return path
    else:
        try:
            import pwd
        except ImportError:
            # pwd module unavailable, return path unchanged
            return path
        name = path[1:i]
        try:
            pwent = pwd.getpwnam(name)
        except KeyError:
            # bpo-10496: if the user name from the path doesn't exist in the
            # password database, return the path unchanged
            return path
        userhome = pwent.pw_dir

    root = "/"
    userhome = userhome.rstrip(root)
    return (userhome + path[i:]) or root


def fix_perms(fn: str, ctx: Context):
    os.chown(fn, ctx.euid, ctx.egid)


def is_steam_gamepad_running(ctx: Context | None, gamepadui: bool = True):
    pid = None
    try:
        with open(expanduser(STEAM_PID, ctx)) as f:
            pid = f.read().strip()

        steam_cmd_path = f"/proc/{pid}/cmdline"
        if not os.path.exists(steam_cmd_path):
            return False

        # The command line is irrelevant if we just want to know if Steam is running.
        if not gamepadui:
            return True

        # Use this and line to determine if Steam is running in DeckUI mode.
        with open(steam_cmd_path, "rb") as f:
            steam_cmd = f.read()
        is_deck_ui = b"-gamepadui" in steam_cmd
        if not is_deck_ui:
            return False
    except Exception:
        return False
    return True


def run_steam_command(command: str, ctx: Context):
    global home_path
    try:
        if ctx.euid != ctx.uid:
            result = subprocess.run(
                [
                    "su",
                    ctx.name,
                    "-c",
                    f"{expanduser(STEAM_EXE, ctx)} -ifrunning {command}",
                ]
            )
        else:
            result = subprocess.run([expanduser(STEAM_EXE, ctx), "-ifrunning", command])

        return result.returncode == 0
    except Exception as e:
        logger.error(f"Received error when running steam command `{command}`\n{e}")
    return False


def open_steam_kbd(emit, open: bool = True):
    return (
        emit
        and is_steam_gamepad_running(emit.ctx, False)
        and run_steam_command(
            f"steam://{'open' if open else 'close'}/keyboard", emit.ctx
        )
    )



import logging
import os
import select
import subprocess
import time
from collections import deque
from typing import Literal

from hhd.controller.base import Consumer, Producer

logger = logging.getLogger(__name__)


def gen_cmd(
    cid: int, cmd: bytes | list[int] | str, idx: int | None = None, size: int = 64
):
    # Command: [idx, cid, 0x3f, *cmd, 0x3f, cid], idx is optional
    if isinstance(cmd, str):
        c = bytes.fromhex(cmd)
    else:
        c = bytes(cmd)
    base = bytes([cid, 0x3F, *c])
    if idx is not None:
        base = bytes([idx]) + base
    return base + bytes([0] * (size - len(base) - 2)) + bytes([0x3F, cid])


def gen_rgb_mode(mode: str):
    mc = 0
    match mode:
        case "monster_woke":
            mc = 0x0D
        case "flowing":
            mc = 0x03
        case "sunset":
            mc = 0x0B
        case "neon":
            mc = 0x05
        case "dreamy":
            mc = 0x07
        case "cyberpunk":
            mc = 0x09
        case "colorful":
            mc = 0x0C
        case "aurora":
            mc = 0x01
        case "sun":
            mc = 0x08
    return gen_cmd(0xFD, [0x00, mc])


gen_intercept = lambda enable: gen_cmd(0xA1, 2 * [int(enable)], idx=0x00)


def gen_brightness(
    side: Literal[0, 3, 4],
    enabled: bool,
    brightness: Literal["low", "medium", "high"],
):
    match brightness:
        case "low":
            bc = 0x01
        case "medium":
            bc = 0x03
        case _:  # "high":
            bc = 0x04

    return gen_cmd(
        0xFD, [side, 0xFD, 0x00 if side else 0x03, 0x00, int(enabled), 0x05, bc]
    )


def gen_rgb_solid(r, g, b, side: Literal[0x00, 0x03, 0x04] = 0x00):
    start = [side, 0xFE, 0x00, 0x00]
    end = [r, g]
    return gen_cmd(0xFD, start + 18 * [r, g, b] + end)


KBD_NAME = "keyboard"
KBD_NAME_NON_TURBO = "share"
KBD_HOLD = 0.12
OXP_BUTTONS = {
    0x24: KBD_NAME,
    0x22: "extra_l1",
    0x23: "extra_r1",
}


INITIALIZE = [
    # gen_intercept(True),
    gen_cmd(
        0xF5,
        "0000000001010101000000020102000000030103000000040104000000050105000000060106000000070107000000080108000000090109000000",
        idx=0x01,
    ),
    gen_cmd(
        0xF5,
        "00000000010a010a0000000b010b0000000c010c0000000d010d0000000e010e0000000f010f000000100110000000220200000000230200000000",
        idx=0x02,
    ),
    # gen_intercept(False), # does not seem to be needed
]

INIT_DELAY = 2
WRITE_DELAY = 0.05
SCAN_DELAY = 1

_mappings_init = True


def get_serial():

    VID = "1a86"
    PID = "7523"

    dev = None
    buttons_only = False
    for d in os.listdir("/dev"):
        if not d.startswith("ttyUSB"):
            continue

        path = os.path.join("/dev", d)

        out = subprocess.run(
            ["udevadm", "info", "--name", path],
            check=True,
            capture_output=True,
            text=True,
        )

        if f"ID_VENDOR_ID={VID}" not in out.stdout:
            continue

        if f"ID_MODEL_ID={PID}" not in out.stdout:
            continue

        dev = path
        break

    for d in os.listdir("/dev"):
        if not d.startswith("ttyS"):
            continue

        path = os.path.join("/dev", d)

        out = subprocess.run(
            ["udevadm", "info", "--name", path],
            check=True,
            capture_output=True,
            text=True,
        )

        # OneXFly device is pnp
        if "devices/pnp" not in out.stdout:
            continue

        # TODO: We need to get a baseline to quirk this type properly
        logger.info(f"Serial port information:\n{out.stdout}")

        dev = path
        buttons_only = True
        break
    return dev, buttons_only


def init_serial():
    import serial

    dev, buttons_only = get_serial()

    if not dev:
        logger.warning("OXP CH340 serial device not found.")
        return None, buttons_only

    logger.info(f"OXP CH340 serial device found at {dev}")

    ser = serial.Serial(
        dev,
        115200,
        parity=serial.PARITY_EVEN,
        stopbits=serial.STOPBITS_TWO,
        bytesize=serial.EIGHTBITS,
        timeout=0,
        exclusive=True,
    )

    return ser, buttons_only


class SerialDevice(Consumer, Producer):
    def __init__(self, required: bool = False, turbo: bool = True) -> None:
        self.required = required
        self.ser = None
        self.buf = bytearray()
        self.prev = {}
        self.queue_kbd = None
        self.queue_cmd = deque(maxlen=10)
        self.last_sent = 0
        self.queue_led = None
        self.turbo = turbo
        self.buttons_only = False

        self.prev_brightness = None
        self.prev_stick = None
        self.prev_stick_enabled = None
        self.prev_center = None
        self.prev_center_enabled = None

    def open(self):
        ser, self.buttons_only = init_serial()
        if ser is None:
            if self.required:
                raise RuntimeError("OXP CH340 serial device not found.")
            return []
        self.ser = ser
        self.queue_kbd = None
        self.prev = {}

        self.next_send = time.perf_counter() + INIT_DELAY

        global _mappings_init
        if _mappings_init:
            self.queue_cmd.extend(INITIALIZE)
            _mappings_init = False

        return [ser.fd]

    def consume(self, events):
        if not self.ser:
            return

        # Capture led events
        for ev in events:
            if ev["type"] == "led":
                # if self.queue_led:
                #     logger.warning("OXP CH340 LED event queue overflow.")
                self.queue_led = ev

        # Send queued event if applicable
        curr = time.perf_counter()
        if self.queue_cmd and curr - self.last_sent > WRITE_DELAY:
            cmd = self.queue_cmd.popleft()
            logger.info(f"OXP C: {cmd.hex()}")
            self.ser.write(cmd)
            self.last_sent = time.perf_counter()
        
        # No LEDs, skip the rest
        if self.buttons_only:
            return

        # Queue needs to flush before switching to next event
        # Also, there needs to be a led event to queue
        if self.queue_cmd or not self.queue_led:
            return
        ev = self.queue_led
        self.queue_led = None

        brightness = "high"
        stick = None
        stick_enabled = True
        center = None
        center_enabled = True
        init = ev["initialize"]

        match ev["mode"]:
            case "solid":
                stick = ev["red"], ev["green"], ev["blue"]
                r2, g2, b2 = ev["red2"], ev["green2"], ev["blue2"]
                center = r2, g2, b2
                center_enabled = r2 > 10 or g2 > 10 or b2 > 10
            case "duality":
                stick = ev["red"], ev["green"], ev["blue"]
                center = ev["red2"], ev["green2"], ev["blue2"]
            case "oxp":
                brightness = ev["brightnessd"]
                stick = ev["oxp"]
                if stick == "classic":
                    # Classic mode is a cherry red
                    stick = 0xB7, 0x30, 0x00
                r2, g2, b2 = ev["red2"], ev["green2"], ev["blue2"]
                center = r2, g2, b2
                center_enabled = r2 > 10 or g2 > 10 or b2 > 10
                init = True
            case _:  # "disabled":
                stick_enabled = False
                center_enabled = False

        # Force RGB to not initialize to workaround RGB breaking
        # rumble when being set
        if self.prev_stick_enabled is None:
            self.prev_stick_enabled = stick_enabled
        if self.prev_brightness is None:
            self.prev_brightness = brightness
        if self.prev_stick is None:
            self.prev_stick = stick

        if (
            stick_enabled != self.prev_stick_enabled
            or brightness != self.prev_brightness
        ):
            self.queue_cmd.append(gen_brightness(0, stick_enabled, brightness))
            self.prev_brightness = brightness
            self.prev_stick_enabled = stick_enabled

        if stick_enabled and stick != self.prev_stick:
            if isinstance(stick, str):
                self.queue_cmd.append(gen_rgb_mode(stick))
            else:
                self.queue_cmd.append(gen_rgb_solid(*stick, side=0x00))
            self.prev_stick = stick
            self.prev_brightness = brightness
            self.prev_stick_enabled = stick_enabled

        if center_enabled != self.prev_center_enabled:
            self.queue_cmd.append(gen_brightness(0x03, center_enabled, "high"))
            self.queue_cmd.append(gen_brightness(0x04, center_enabled, "high"))
            self.prev_center_enabled = center_enabled

        # Only apply center colors on init on init
        if init and center_enabled and center and center != self.prev_center:
            self.queue_cmd.append(gen_rgb_solid(*center, side=0x03))
            self.queue_cmd.append(gen_rgb_solid(*center, side=0x04))
            self.prev_center = center

    def produce(self, fds):
        if not self.ser:
            return []

        evs = []
        if self.queue_kbd:
            curr = time.perf_counter()
            if curr - KBD_HOLD > self.queue_kbd:
                evs = [
                    {
                        "type": "button",
                        "code": KBD_NAME if self.turbo else KBD_NAME_NON_TURBO,
                        "value": False,
                    }
                ]
                self.queue_kbd = None

        if self.ser.fd not in fds:
            return evs

        CMD_LEN = 14

        while out := self.ser.read():
            self.buf.extend(out)

        while len(self.buf) >= CMD_LEN:
            # Align to start boundary
            if self.buf[1] != 0x3F:
                self.buf = self.buf[1:]
                continue

            # Grab command id
            cmd = self.buf[:CMD_LEN]
            self.buf = self.buf[CMD_LEN:]
            cid = cmd[0]

            valid = cmd[-2] == 0x3F and cmd[-1] == cid
            if not valid:
                logger.warning(f"OXP CH340 invalid command: {self.buf.hex()}")
                continue

            if cid == 0xEF:
                # Initialization command, skip
                continue

            if cid != 0x1A:
                logger.warning(f"OXP CH340 unknown command: {cmd.hex()}")
                continue

            btn = cmd[2]

            if btn not in OXP_BUTTONS:
                logger.warning(
                    f"OXP CH340 unknown button: {btn:x} from cmd:\n{cmd.hex()}"
                )
                continue

            btn = OXP_BUTTONS[btn]
            pressed = cmd[8] == 1

            if btn == KBD_NAME:
                if pressed and (btn not in self.prev or self.prev[btn] != pressed):
                    evs.append(
                        {
                            "type": "button",
                            "code": KBD_NAME if self.turbo else KBD_NAME_NON_TURBO,
                            "value": True,
                        }
                    )
                    self.queue_kbd = time.perf_counter()
                self.prev[btn] = pressed
                continue

            # logger.info(f"OXP CH340 button: {btn} pressed: {pressed}")
            if btn in self.prev and self.prev[btn] == pressed:
                # Debounce
                continue

            self.prev[btn] = pressed
            evs.append(
                {
                    "type": "button",
                    "code": btn,
                    "value": pressed,
                }
            )

        return evs



import json
import logging
import os
import select
import signal
import subprocess
import shutil
import time
from threading import Lock, Thread
from typing import Literal, Sequence

from hhd.i18n import _
from hhd.plugins import Context, HHDPlugin, HHDSettings, load_relative_yaml
from hhd.plugins.conf import Config

logger = logging.getLogger(__name__)

REFRESH_HZ = 3
PROGRESS_STAGES = {
    "pulling": (_("Downloading:"), 0, 80),
    "importing": (_("Importing:"), 80, 10),
    "staging": (_("Deploying:"), 90, 10),
    "unknown": (_("Loading"), 100, 0),
}

BOOTC_ENABLED = os.environ.get("HHD_BOOTC", "0") == "1"
BOOTC_PATH = os.environ.get("HHD_BOOTC_PATH", "bootc")
BRANCHES = os.environ.get(
    "HHD_BOOTC_BRANCHES", "stable:Stable,testing:Testing,unstable:Unstable"
)

REF_PREFIX = "§ "
DEFAULT_PREFIX = "◉ "

BOOTC_STATUS_CMD = [
    BOOTC_PATH,
    "status",
    "--format",
    "json",
]

RPM_OSTREE_RESET = [
    "rpm-ostree",
    "reset",
]

BOOTC_CHECK_CMD = [
    BOOTC_PATH,
    "update",
    "--check",
]

BOOTC_ROLLBACKCMD = [
    BOOTC_PATH,
    "rollback",
]

BOOTC_UPDATE_CMD = [
    BOOTC_PATH,
    "update",
]

SKOPEO_REBASE_CMD = lambda ref: ["skopeo", "inspect", "docker://" + ref]


STAGES = Literal[
    "init",
    "ready",
    "ready_check",
    "ready_updated",
    "ready_reverted",
    "ready_rebased",
    "incompatible",
    "rebase_dialog",
    "loading",
    "loading_rebase",
    "loading_cancellable",
]


def get_bootc_status():
    try:
        output = subprocess.check_output(BOOTC_STATUS_CMD).decode("utf-8")
        return json.loads(output)
    except Exception as e:
        logger.error(f"Failed to get bootc status: {e}")
        return {}


def get_ref_from_status(status: dict | None):
    return (((status or {}).get("spec", None) or {}).get("image", None) or {}).get(
        "image", ""
    )


def get_branch(ref: str, branches: dict, fallback: bool = True):
    if ":" not in ref:
        return next(iter(branches))
    curr_tag = ref[ref.rindex(":") + 1 :]

    for branch in branches:
        if branch in curr_tag:
            return branch

    if not fallback:
        return None
    # If no tag, assume it is the first one
    return next(iter(branches))


def get_rebase_refs(ref: str, tags, lim: int = 7, branches: dict = {}):
    logger.info(f"Getting rebase refs for {ref}")
    try:
        output = subprocess.check_output(SKOPEO_REBASE_CMD(ref)).decode("utf-8")
        data = json.loads(output)
        versions = data.get("RepoTags", [])

        for branch in branches:
            same_branch = [v for v in versions if v.startswith(branch) and v != branch]
            same_branch.sort(reverse=True)
            tags[branch] = same_branch[:lim]

        logger.info(f"Finished getting refs")
    except Exception as e:
        logger.error(f"Failed to get rebase refs: {e}")


def run_command_threaded(cmd: list):
    try:
        return subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
        )
    except Exception as e:
        logger.error(f"Failed to run command: {e}")


def _bootc_progress_reader(fd, emit, friendly, lock, obj):
    last_update = 0
    try:
        while select.select([fd.fileno()], [], [])[0]:
            data = fd.readline()
            if not data:
                break
            data = json.loads(data)

            text, start, length = PROGRESS_STAGES.get(
                data.get("task", "unknown"), PROGRESS_STAGES["unknown"]
            )

            match data["type"]:
                case "ProgressSteps":
                    curr = data.get("steps", 0)
                    total = data.get("stepsTotal", 0)
                    value = start + min(length, int((curr / (total + 1)) * length))
                    if total > 1:
                        unit = f" {friendly} ({min(curr + 1, total)}/{total})"
                    else:
                        unit = f" {friendly}"
                case "ProgressBytes":
                    curr = data.get("bytes", 0)
                    total = data.get("bytesTotal", 0)
                    value = start + min(length, int((curr / total) * length))
                    unit = f" {friendly} ({curr/1e9:.1f}/{total/1e9 + 0.099:.1f} GB)"
                case _:
                    value = None  # indeterminate

            with lock:
                obj.update({"text": text, "value": value, "unit": unit})

            # Increase the update rate of the UI
            curr = time.perf_counter()
            if curr - last_update > 1 / REFRESH_HZ and emit:
                last_update = curr
                emit({"type": "special", "event": "refresh"})
    finally:
        fd.close()


def run_command_threaded_progress(cmd: list, emit, friendly, lock):
    r = None
    try:
        r, w = os.pipe2(0)
        proc = subprocess.Popen(
            cmd + [f"--json-fd", str(w), "--quiet"],
            pass_fds=[w],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        os.close(w)

        obj = {}
        fd = os.fdopen(r, "r")
        t = Thread(target=_bootc_progress_reader, args=(fd, emit, friendly, lock, obj))
        t.start()
        return proc, obj
    except Exception as e:
        logger.error(f"Failed to run command: {e}")
        if r:
            os.close(r)
        return None, None


def is_incompatible(status: dict):
    if status.get("apiVersion", None) != "org.containers.bootc/v1":
        return True

    boot_incompatible = (
        (status.get("status", None) or {}).get("booted", None) or {}
    ).get("incompatible", False)

    if staged := ((status.get("status", None) or {}).get("staged", None) or {}):
        return staged.get("incompatible", False)

    return boot_incompatible


def has_bootc_progress_support():
    return "--json-fd" in subprocess.check_output(
        [BOOTC_PATH, "upgrade", "--help"]
    ).decode("utf-8")


class BootcPlugin(HHDPlugin):
    def __init__(self) -> None:
        self.name = f"bootc"
        self.priority = 70
        self.log = "bupd"
        self.proc = None
        self.branch_name = None
        self.branch_ref = None
        self.checked_update = False
        self.t = None
        self.t_data = None
        self.progress_lock = Lock()
        self.progress = None
        self.staged = ""
        self.cached_version = ""
        self.emit = None

        self.branches = {}
        for branch in BRANCHES.split(","):
            name, display = branch.split(":")
            self.branches[name] = display

        self.status = None
        self.enabled = True
        self.state: STAGES = "init"

    def settings(self) -> HHDSettings:
        if self.enabled:
            sets = {"updates": {"bootc": load_relative_yaml("settings.yml")}}

            sets["updates"]["bootc"]["children"]["stage"]["modes"]["rebase"][
                "children"
            ]["branch"]["options"] = self.branches

            return sets
        else:
            return {}

    def open(
        self,
        emit,
        context: Context,
    ):
        self.updated = False
        self.bootc_progress = has_bootc_progress_support()
        self.emit = emit
        if self.bootc_progress:
            logger.info("Bootc progress support detected")
        else:
            logger.warning("Bootc progress support not detected")

    def get_version(self, s):
        assert self.status
        return (
            (self.status.get("status", {}).get(s, None) or {}).get("image", None) or {}
        ).get("version", "")

    def _init(self, conf: Config):
        self.status = get_bootc_status()

        if is_incompatible(self.status):
            conf["updates.bootc.stage.mode"] = "incompatible"
            self.state = "incompatible"
            conf[f"updates.bootc.update"] = None
            conf[f"updates.bootc.steamos-update"] = "incompatible"
            return

        ref = ((self.status.get("spec", None) or {}).get("image", None) or {}).get(
            "image", ""
        )
        img = ref
        if "/" in img:
            img = img[img.rfind("/") + 1 :]

        # Find branch and replace tag
        branch = get_branch(img, self.branches)
        rebased_ver = None
        self.branch_name = branch
        self.branch_ref = None
        has_rebased = False
        if branch:
            if ":" in img:
                tag = img[img.rindex(":") + 1 :]
                if tag != branch:
                    rebased_ver = tag
                    has_rebased = True
                    self.branch_ref = ref.split(":")[0] + ":" + branch
                img = img[: img.rindex(":") + 1] + branch
        if img:
            conf["updates.bootc.image"] = img

        # If we have a staged update, that will boot first
        self.staged = og = s = self.get_version("staged")
        staged = False
        if s:
            s = DEFAULT_PREFIX + s
            staged = True
        if s and rebased_ver and og in rebased_ver:
            s = REF_PREFIX + s
            # Only apply one start to avoid confusion
            rebased_ver = None
        conf["updates.bootc.staged"] = s

        # Check if the user selected rollback
        # Then that will be the default, provided there is a rollback
        rollback = (
            not staged
            and (self.status.get("spec", None) or {}).get("bootOrder", None)
            == "rollback"
        )
        s = self.get_version("rollback")
        if s and rollback:
            s = DEFAULT_PREFIX + s
        else:
            rollback = False
        conf[f"updates.bootc.rollback"] = s

        # Otherwise, the booted version will be the default
        og = s = self.get_version("booted")
        if s and not rollback and not staged:
            s = DEFAULT_PREFIX + s
        if s and rebased_ver and og in rebased_ver:
            s = REF_PREFIX + s
        conf[f"updates.bootc.booted"] = s

        conf["updates.bootc.status"] = ""
        self.updated = True

        cached = self.status.get("status", {}).get("booted", {}).get("cachedUpdate", {})
        cached_version = cached.get("version", "") if cached else ""
        cached_img = cached.get("image", {}).get("image", "") if cached else ""
        if "/" in cached_img:
            cached_img = cached_img[cached_img.rfind("/") + 1 :]
        self.cached_version = cached_version

        if self.checked_update:
            conf[f"updates.bootc.update"] = _("No update available")
        else:
            conf[f"updates.bootc.update"] = None

        if (
            cached_version
            and cached_img == img
            and cached_version != self.get_version("staged")
        ):
            conf["updates.bootc.stage.mode"] = "ready"
            self.state = "ready"
            conf[f"updates.bootc.update"] = cached_version
            conf[f"updates.bootc.steamos-update"] = "has-update"
            conf["updates.bootc.steamos-target"] = None
        elif self.get_version("staged"):
            conf["updates.bootc.stage.mode"] = "ready_updated"
            self.state = "ready_updated"
            conf[f"updates.bootc.steamos-update"] = "updated"
        elif has_rebased:
            conf["updates.bootc.stage.mode"] = "ready_rebased"
            self.state = "ready_rebased"
            conf[f"updates.bootc.steamos-update"] = "updated"
        elif rollback:
            conf["updates.bootc.stage.mode"] = "ready_reverted"
            self.state = "ready_reverted"
            conf[f"updates.bootc.steamos-update"] = "updated"
        else:
            conf["updates.bootc.stage.mode"] = "ready_check"
            self.state = "ready_check"
            conf[f"updates.bootc.steamos-update"] = "ready"

    def update(self, conf: Config):

        # Detect reset and avoid breaking the UI
        if conf.get("updates.bootc.stage.mode", None) is None:
            self._init(conf)
            return

        # Try to fill in basic info
        match self.state:
            case "init":
                self._init(conf)
            # Ready
            case (
                "ready"
                | "ready_check"
                | "ready_updated"
                | "ready_reverted"
                | "ready_rebased" as e
            ):
                update = conf.get_action(f"updates.bootc.stage.{e}.update")
                revert = conf.get_action(f"updates.bootc.stage.{e}.revert")
                rebase = conf.get_action(f"updates.bootc.stage.{e}.rebase")
                reboot = conf.get_action(f"updates.bootc.stage.{e}.reboot")

                steamos = conf.get("updates.bootc.steamos-update", None)
                target = conf.get("updates.bootc.steamos-target", None)
                if target and (target not in BRANCHES or target == self.branch_name):
                    target = None

                # Handle steamos polkit
                steamos_upd = False
                if steamos == "check":
                    if target or e == "ready":
                        conf["updates.bootc.steamos-update"] = "has-update"
                    else:
                        update = True
                        steamos_upd = True
                if steamos == "apply":
                    if target or e == "ready":
                        update = True
                        steamos_upd = True

                if steamos_upd and target:
                    curr = get_ref_from_status(self.status)
                    next_ref = (
                        (curr[: curr.rindex(":")] if ":" in curr else curr)
                        + ":"
                        + target
                    )
                    if next_ref == curr:
                        conf["updates.bootc.steamos-target"] = None
                        self._init(conf)
                    self.state = "loading_cancellable"
                    cmd = [BOOTC_PATH, "switch", next_ref]
                    if self.bootc_progress:
                        self.proc, self.progress = run_command_threaded_progress(
                            cmd, self.emit, target, self.progress_lock
                        )
                    else:
                        self.proc = run_command_threaded(cmd)
                    conf["updates.bootc.stage.mode"] = "loading_cancellable"
                    conf["updates.bootc.stage.loading_cancellable.progress"] = {
                        "text": _("Rebasing to "),
                        "unit": self.branches.get(target, target),
                        "value": None,
                    }
                elif update:
                    if e == "ready_rebased" and self.branch_ref:
                        self.checked_update = False
                        self.state = "loading_cancellable"
                        cmd = [BOOTC_PATH, "switch", self.branch_ref]
                        if self.bootc_progress:
                            self.proc, self.progress = run_command_threaded_progress(
                                cmd,
                                self.emit,
                                self.branch_name,
                                self.progress_lock,
                            )
                        else:
                            self.proc = run_command_threaded(cmd)
                        conf["updates.bootc.stage.mode"] = "loading_cancellable"
                        conf["updates.bootc.stage.loading_cancellable.progress"] = {
                            "text": _("Updating to latest "),
                            "unit": self.branches.get(
                                self.branch_name, self.branch_name
                            ),
                            "value": None,
                        }
                    elif e == "ready":
                        self.state = "loading_cancellable"
                        self.checked_update = False
                        if self.bootc_progress:
                            self.proc, self.progress = run_command_threaded_progress(
                                BOOTC_UPDATE_CMD,
                                self.emit,
                                self.cached_version or self.branch_name or "",
                                self.progress_lock,
                            )
                        else:
                            self.proc = run_command_threaded(BOOTC_UPDATE_CMD)
                        conf["updates.bootc.stage.mode"] = "loading_cancellable"
                        conf["updates.bootc.stage.loading_cancellable.progress"] = {
                            "text": _("Updating... "),
                            "value": None,
                            "unit": None,
                        }
                    else:
                        self.state = "loading"
                        self.proc = run_command_threaded(BOOTC_CHECK_CMD)
                        self.checked_update = True
                        conf["updates.bootc.stage.mode"] = "loading"
                        conf["updates.bootc.stage.loading.progress"] = {
                            "text": _("Checking for updates..."),
                            "value": None,
                            "unit": None,
                        }
                elif revert:
                    self.checked_update = False
                    self.state = "loading"
                    self.proc = run_command_threaded(BOOTC_ROLLBACKCMD)
                    conf["updates.bootc.stage.mode"] = "loading"
                    if e == "ready_updated":
                        text = _("Undoing Update...")
                    elif e == "ready_reverted":
                        text = _("Undoing Revert...")
                    else:
                        text = _("Reverting to Previous version...")
                    conf["updates.bootc.stage.loading.progress"] = {
                        "text": text,
                        "value": None,
                        "unit": None,
                    }
                elif rebase:
                    self.checked_update = False
                    if not self.branches:
                        self._init(conf)
                    else:
                        # Get branch that should be default
                        curr = (
                            (self.status or {})
                            .get("spec", {})
                            .get("image", {})
                            .get("image", "")
                        )
                        default = get_branch(curr, self.branches)
                        conf["updates.bootc.stage.rebase.branch"] = default

                        # Prepare loader
                        conf["updates.bootc.stage.mode"] = "loading"
                        conf["updates.bootc.stage.loading.progress"] = {
                            "text": _("Loading Versions..."),
                            "value": None,
                            "unit": None,
                        }

                        # Launch loader thread
                        self.t_data = {}
                        self.t = Thread(
                            target=get_rebase_refs,
                            args=(curr, self.t_data),
                            kwargs={"branches": self.branches},
                        )
                        self.t.start()
                        self.state = "loading_rebase"
                elif reboot:
                    logger.info("User pressed reboot in updater. Rebooting...")
                    subprocess.run(["systemctl", "reboot"])

            # Incompatible
            case "incompatible":
                if conf.get_action("updates.bootc.stage.incompatible.reset"):
                    self.state = "loading"
                    self.proc = run_command_threaded(RPM_OSTREE_RESET)
                    conf["updates.bootc.stage.mode"] = "loading"
                    conf["updates.bootc.stage.loading.progress"] = {
                        "text": _("Removing Customizations..."),
                        "value": None,
                        "unit": None,
                    }

            # Rebase dialog
            case "rebase_dialog" | "loading_rebase" as e:
                # FIXME: this is the only match statement that
                # does early returns. Allows loading the previous
                # versions instantly.

                conf["updates.bootc.update"] = None
                if e == "loading_rebase":
                    if self.t is None:
                        self._init(conf)
                        return
                    elif not self.t.is_alive():
                        self.t = None
                        self.state = "rebase_dialog"
                        conf["updates.bootc.stage.mode"] = "rebase"
                    else:
                        return

                apply = conf.get_action("updates.bootc.stage.rebase.apply")
                cancel = conf.get_action("updates.bootc.stage.rebase.cancel")
                branch = conf.get(
                    "updates.bootc.stage.rebase.branch", next(iter(self.branches))
                )

                version = "latest"
                if not self.t_data:
                    conf["updates.bootc.stage.rebase.version_error"] = _(
                        "Failed to load previous versions"
                    )
                else:
                    conf["updates.bootc.stage.rebase.version_error"] = None
                    if branch in self.t_data:
                        bdata = {k.replace(".", ""): k for k in self.t_data[branch]}
                        version = conf.get(
                            "updates.bootc.stage.rebase.version.value", "latest"
                        )
                        conf["updates.bootc.stage.rebase.version"] = None
                        conf["updates.bootc.stage.rebase.version"] = {
                            "options": {
                                "latest": "Latest",
                                **bdata,
                            },
                            "value": version if version in bdata else "latest",
                        }
                        # Readd . since config system does not support them
                        version = bdata.get(version, "latest")

                if cancel:
                    self._init(conf)
                elif apply:
                    if version == "latest":
                        version = branch

                    curr = get_ref_from_status(self.status)
                    next_ref = (
                        (curr[: curr.rindex(":")] if ":" in curr else curr)
                        + ":"
                        + version
                    )
                    if next_ref == curr:
                        self._init(conf)
                    else:
                        self.state = "loading_cancellable"
                        cmd = [BOOTC_PATH, "switch", next_ref]
                        if self.bootc_progress:
                            self.proc, self.progress = run_command_threaded_progress(
                                cmd, self.emit, version, self.progress_lock
                            )
                        else:
                            self.proc = run_command_threaded(cmd)
                        conf["updates.bootc.stage.mode"] = "loading_cancellable"
                        conf["updates.bootc.stage.loading_cancellable.progress"] = {
                            "text": _("Rebasing to "),
                            "unit": self.branches.get(version, version),
                            "value": None,
                        }

            # Wait for the subcommand to complete
            case "loading_cancellable":
                cancel = conf.get_action(
                    f"updates.bootc.stage.loading_cancellable.cancel"
                )
                if self.proc is None:
                    self._init(conf)
                elif self.proc.poll() is not None:
                    self._init(conf)
                    self.proc = None
                elif cancel:
                    logger.info("User cancelled update. Stopping...")
                    self.proc.send_signal(signal.SIGINT)
                    self.proc.wait()
                    self.proc = None
                    self._init(conf)
                elif self.progress:
                    with self.progress_lock:
                        conf["updates.bootc.stage.loading_cancellable.progress"] = (
                            self.progress
                        )
                        val = self.progress.get("value", None)
                        if val is not None:
                            try:
                                val = int(val)
                                conf["updates.bootc.steamos-update"] = f"{val}%"
                            except ValueError:
                                pass
            case "loading":
                if self.proc is None:
                    self._init(conf)
                elif self.proc.poll() is not None:
                    self._init(conf)
                    self.proc = None

    def close(self):
        if self.proc:
            self.proc.send_signal(signal.SIGINT)
            self.proc.wait()
            self.proc = None
        if self.t:
            if self.t.is_alive():
                self.t.join()
            self.t = None


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    if not BOOTC_ENABLED:
        return []
    
    if not shutil.which(BOOTC_PATH):
        logger.warning("Bootc is enabled but not found in path.")
        return []

    return [BootcPlugin()]



from typing import Sequence

from hhd.plugins import (
    HHDPlugin,
)

from .slim import LegionGoSControllerPlugin
from .tablet import LegionGoControllersPlugin

LEGION_GO_CONFS = {
    "83E1": {
        "name": "Legion Go",
    },
}

LEGION_S_CONFS = {
    "83L3": {
        "name": "Legion Go S Z2 Go",
    },
    "83N6": {
        "name": "Legion Go S Z1E",
    },
    "83Q2": {
        "name": "Legion Go S",
    },
    "83Q3": {
        "name": "Legion Go S",
    },
}


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    # Match just product number, should be enough for now
    with open("/sys/devices/virtual/dmi/id/product_name") as f:
        dmi = f.read().strip()

    if dmi in LEGION_S_CONFS:
        return [LegionGoSControllerPlugin(dconf=LEGION_S_CONFS[dmi])]

    if dmi in LEGION_GO_CONFS:
        return [LegionGoControllersPlugin(dconf=LEGION_GO_CONFS[dmi])]

    return []



import logging
import os
import select
import subprocess
from threading import Event
from time import perf_counter, sleep
from typing import cast

import evdev

from hhd.utils import Context, is_steam_gamepad_running, run_steam_command

from .const import PowerButtonConfig
from ..power.power import supports_sleep, emergency_hibernate

logger = logging.getLogger(__name__)

STEAM_WAIT_DELAY = 0.5
LONG_PRESS_DELAY = 2.0
DEBOUNCE_DELAY = 1


def B(b: str):
    return cast(int, getattr(evdev.ecodes, b))


def register_power_buttons(b: PowerButtonConfig) -> list[evdev.InputDevice]:
    out = []
    for device in [evdev.InputDevice(path) for path in evdev.list_devices()]:
        capture = False
        for phys in b.phys:
            if str(device.phys).startswith(phys):
                capture = True
        if capture:
            device.grab()
            logger.info(f"Captured power button '{device.name}': '{device.phys}'")
            out.append(device)
    return out


def pick_closest_button(btns: list[evdev.InputDevice], cfg: PowerButtonConfig):
    for phys in cfg.phys:
        for b in btns:
            if str(b.phys).startswith(phys):
                return b

    if btns:
        return btns[0]
    return None


def register_hold_button(b: PowerButtonConfig) -> evdev.InputDevice | None:
    if not b.hold_phys or not b.hold_code:
        logger.error(
            f"Device configuration tuple does not contain required parameters:\n{b}"
        )
        return None

    for device in [evdev.InputDevice(path) for path in evdev.list_devices()]:
        for phys in b.hold_phys:
            if str(device.phys).startswith(phys):
                if b.hold_grab:
                    device.grab()
                logger.info(f"Captured hold keyboard '{device.name}': '{device.phys}'")
                return device
    return None


_supports_sleep = None


def run_steam_shortpress(perms: Context):
    global _supports_sleep
    # FIXME: This should be patched in systemd instead
    if _supports_sleep is None:
        _supports_sleep = supports_sleep()

    if _supports_sleep:
        return run_steam_command("steam://shortpowerpress", perms)
    else:
        emergency_hibernate(shutdown=False)
        return True


def run_steam_longpress(perms: Context):
    return run_steam_command("steam://longpowerpress", perms)


def power_button_run(cfg: PowerButtonConfig, ctx: Context, should_exit: Event, emit):
    match cfg.type:
        case "only_press":
            logger.info(
                f"Starting multi-device powerbutton handler for device '{cfg.device}'."
            )
            power_button_multidev(cfg, ctx, should_exit, emit)
        case "hold_emitted":
            logger.info(
                f"Starting timer based powerbutton handler for device '{cfg.device}'."
            )
            power_button_timer(cfg, ctx, should_exit, emit)
        case "hold_isa":
            logger.info(
                f"Starting isa keyboard powerbutton handler for device '{cfg.device}'."
            )
            power_button_isa(cfg, ctx, should_exit, emit)
        case _:
            logger.error(f"Invalid type in config '{cfg.type}'. Exiting.")


def power_button_isa(cfg: PowerButtonConfig, perms: Context, should_exit: Event, emit):
    press_dev = None
    press_devs = []
    hold_dev = None
    try:
        while not should_exit.is_set():
            # Initial check for steam
            if not is_steam_gamepad_running(perms):
                # Close devices
                if press_devs:
                    for d in press_devs:
                        d.close()
                    press_devs = []
                if press_dev:
                    press_dev.close()
                    press_dev = None
                if hold_dev:
                    hold_dev.close()
                    hold_dev = None
                logger.info(f"Waiting for steam to launch.")
                while not is_steam_gamepad_running(perms):
                    if should_exit.is_set():
                        return
                    sleep(STEAM_WAIT_DELAY)

            if not press_dev or not hold_dev:
                logger.info(f"Steam is running, hooking power button.")
                press_devs = register_power_buttons(cfg)
                press_dev = press_devs[0] if press_devs else None
                hold_dev = register_hold_button(cfg)
            if not press_dev:
                logger.error(f"Power button interfaces not found, disabling plugin.")
                return

            # Add timeout to release the button if steam exits.
            r = select.select(
                [press_dev.fd, hold_dev.fd] if hold_dev else [press_dev.fd],
                [],
                [],
                STEAM_WAIT_DELAY,
            )[0]

            if not r:
                continue
            fd = r[0]  # handle one button at a time

            # Handle button event
            issue_systemctl = False
            if fd == press_dev.fd:
                ev = press_dev.read_one()
                if ev.type == B("EV_KEY") and ev.code == B("KEY_POWER") and ev.value:
                    logger.info("Executing short press.")
                    issue_systemctl = not run_steam_shortpress(perms)
                    emit({"type": "special", "event": "pbtn_short"})
            elif hold_dev and fd == hold_dev.fd:
                ev = hold_dev.read_one()
                if ev.type == B("EV_KEY") and ev.code == cfg.hold_code and ev.value:
                    logger.info("Executing long press.")
                    issue_systemctl = not run_steam_longpress(perms)
                    emit({"type": "special", "event": "pbtn_long"})

            if issue_systemctl:
                logger.error(
                    "Power button action did not work. Calling `systemctl suspend`"
                )
                os.system("systemctl suspend")
    except KeyboardInterrupt:
        pass
    except Exception as e:
        logger.error(f"Received exception, exitting:\n{e}")


def power_button_timer(
    cfg: PowerButtonConfig, perms: Context, should_exit: Event, emit
):
    dev = None
    devs = []
    try:
        pressed_time = None
        while not should_exit.is_set():
            # Initial check for steam
            if not is_steam_gamepad_running(perms):
                # Close devices
                if devs:
                    for d in devs:
                        d.close()
                        devs = []
                    if dev:
                        dev.close()
                        dev = None
                logger.info(f"Waiting for steam to launch.")
                while not is_steam_gamepad_running(perms):
                    if should_exit.is_set():
                        return
                    sleep(STEAM_WAIT_DELAY)

            if not dev:
                logger.info(f"Steam is running, hooking power button.")
                devs = register_power_buttons(cfg)
                dev = pick_closest_button(devs, cfg)
            if not dev:
                logger.error(f"Power button not found, disabling plugin.")
                return

            # Add timeout to release the button if steam exits.
            delay = LONG_PRESS_DELAY if pressed_time else STEAM_WAIT_DELAY
            r = select.select([dev.fd], [], [], delay)[0]

            # Handle press logic
            if r:
                # Handle button event
                ev = dev.read_one()
                if ev.type == B("EV_KEY") and ev.code == B("KEY_POWER"):
                    curr_time = perf_counter()
                    if ev.value:
                        pressed_time = curr_time
                        press_type = "initial_press"
                    elif pressed_time:
                        if curr_time - pressed_time > LONG_PRESS_DELAY:
                            press_type = "long_press"
                        else:
                            press_type = "short_press"
                        pressed_time = None
                    else:
                        press_type = "release_without_press"
                else:
                    press_type = "no_press"
            elif pressed_time:
                # Button was pressed but we hit a timeout, that means
                # it is a long press
                press_type = "long_press"
                pressed_time = None
            else:
                # Otherwise, no press
                press_type = "no_press"

            issue_systemctl = False
            match press_type:
                case "long_press":
                    logger.info("Executing long press.")
                    issue_systemctl = not run_steam_longpress(perms)
                    emit({"type": "special", "event": "pbtn_long"})
                case "short_press":
                    logger.info("Executing short press.")
                    issue_systemctl = not run_steam_shortpress(perms)
                    emit({"type": "special", "event": "pbtn_short"})
                case "initial_press":
                    logger.info("Power button pressed down.")
                case "release_without_press":
                    logger.error("Button released without being pressed.")

            if issue_systemctl:
                logger.error(
                    "Power button action did not work. Calling `systemctl suspend`"
                )
                os.system("systemctl suspend")
    except KeyboardInterrupt:
        pass
    except Exception as e:
        logger.error(f"Received exception, exitting:\n{e}")
    finally:
        if dev:
            dev.close()
        if devs:
            for d in devs:
                d.close()


def power_button_multidev(
    cfg: PowerButtonConfig, perms: Context, should_exit: Event, emit
):
    devs = []
    fds = []
    last_pressed = None
    try:
        while not should_exit.is_set():
            # Initial check for steam
            if not is_steam_gamepad_running(perms):
                for d in devs:
                    d.close()
                devs = []
                fds = []
                logger.info(f"Waiting for steam to launch.")
                while not is_steam_gamepad_running(perms):
                    if should_exit.is_set():
                        return
                    sleep(STEAM_WAIT_DELAY)

            if not devs:
                logger.info(f"Steam is running, hooking power button.")
                devs = register_power_buttons(cfg)
                fds = {d.fd: d for d in devs}
            if not devs:
                logger.error(f"Power button(s) not found, disabling plugin.")
                return

            # Add timeout to release the button if steam exits.
            r = select.select(list(fds), [], [], STEAM_WAIT_DELAY)[0]

            # Handle press logic
            issue_power = False
            issue_systemctl = False
            for fd in r:
                # Handle button event
                ev = fds[fd].read_one()
                if (
                    ev.type == B("EV_KEY")
                    and ev.code == B("KEY_POWER")
                    and ev.value == 1
                ):
                    curr_time = perf_counter()
                    if not last_pressed or curr_time - last_pressed > DEBOUNCE_DELAY:
                        last_pressed = curr_time
                        issue_power = True

            if issue_power:
                logger.info("Executing short press.")
                issue_systemctl = not run_steam_shortpress(perms)
                emit({"type": "special", "event": "pbtn_short"})

            if issue_systemctl:
                logger.error(
                    "Power button action did not work. Calling `systemctl suspend`"
                )
                os.system("systemctl suspend")
    except KeyboardInterrupt:
        pass
    except Exception as e:
        logger.error(f"Received exception, exitting:\n{e}")
    finally:
        if devs:
            for d in devs:
                d.close()



import logging
from typing import Sequence
import os
import time

from hhd.plugins import Config, Context, HHDPlugin, load_relative_yaml
from .power import (
    get_windows_bootnum,
    boot_windows,
    emergency_hibernate,
    emergency_shutdown,
    delete_temporary_swap,
)

logger = logging.getLogger(__name__)

TEMP_CHECK_INTERVAL = 10
# Chill for the first 10 minutes to avoid bricking installs if a device
# has e.g., a battery bug that trips the condition incorrectly
TEMP_CHECK_INITIALIZE = 300
BATTERY_LOW_THRESHOLD = 5
LAST_ATTEMPT_WAIT = 5
LAST_ATTEMPT_BAIL = 30


def thermal_check(therm: dict[str, int], bat: str | None, last_attempt: float = 0, wakeup: bool = False):
    found = False
    for path, temp in therm.items():
        with open(path) as f:
            curr = int(f.read())
            if curr >= temp:
                logger.warning(
                    f"Thermal zone {path} reached {curr // 1000}C, hibernating."
                )
                found = True

    if bat and not found:
        with open(bat + "/status") as f:
            dc = "discharging" in bat.lower()

        with open(bat + "/capacity") as f:
            curr = int(f.read())
            if dc and curr <= BATTERY_LOW_THRESHOLD:
                logger.warning(f"Battery level reached {curr}%, hibernating.")
                found = True

    if not found:
        return False

    if not wakeup and time.time() - last_attempt < LAST_ATTEMPT_WAIT:
        # There is a small chance that systemctl returns control too early
        # and we run the event loop and fall in the if statement below.
        # Therefore, unless this was triggered by a  wakeup, we should
        # wait a bit.
        return False
    elif time.time() - last_attempt < LAST_ATTEMPT_BAIL:
        # Bail out if we woke up too soon
        # This is to avoid a loop of hibernation attempts and wakeup
        # Hibernation requires ~20s to complete, then boot another 20
        # so a user should not be able to trigger this by waking up
        emergency_shutdown()
        return True
    else:
        emergency_hibernate(shutdown=True)
        return True


def set_bat_alarm(bat: str | None):
    # If we do not do this, systemd will not and the system will wakeup
    # randomly. Systemd only uses the alarm in sleep-then-hibernate and
    # leaves it dangling otherwise.
    # Only touch it if the setting is enabled.
    if not bat or not os.path.exists(bat + "/alarm"):
        return

    if os.path.exists(bat + "/energy_full"):
        with open(bat + "/energy_full") as f:
            full = int(f.read())
    else:
        with open(bat + "/charge_full") as f:
            full = int(f.read())

    # Go a bit below the threshold to make sure we hibernate when we wake up.
    lvl = BATTERY_LOW_THRESHOLD * 85 * full // 100 // 100
    logger.warning(
        f"Setting battery alarm to {lvl}/{full} mAh/mWh ({BATTERY_LOW_THRESHOLD}%)"
    )

    with open(bat + "/alarm", "w") as f:
        f.write(str(lvl))


class PowerPlugin(HHDPlugin):
    def __init__(self) -> None:
        self.name = f"power"
        self.priority = 50
        self.log = "powr"
        self.win_bootnum = None
        self.win_bootnum = get_windows_bootnum()
        self.therm = {}
        self.init = 0
        self.last_check = 0
        self.check_thermal = False
        self.bat = None
        self.alarm_set = False
        self.last_attempt = 0

    def open(
        self,
        emit,
        context: Context,
    ):
        self.started = False
        self.context = context
        self.emit = emit

        self.init = time.time()
        self.therm = {}
        self.bat = None

        try:
            for therm in os.listdir("/sys/class/thermal"):
                if not therm.startswith("thermal_zone"):
                    continue

                with open(f"/sys/class/thermal/{therm}/type") as f:
                    if "acpitz" not in f.read():
                        continue

                for trip in os.listdir(f"/sys/class/thermal/{therm}"):
                    if not trip.startswith("trip_point_"):
                        continue

                    if not trip.endswith("_type"):
                        continue

                    with open(f"/sys/class/thermal/{therm}/{trip}") as f:
                        if "hot" not in f.read():
                            continue

                    with open(
                        f"/sys/class/thermal/{therm}/{trip.replace("_type", "_temp")}"
                    ) as f:
                        self.therm[f"/sys/class/thermal/{therm}/temp"] = int(f.read())
                        break

            for bat in os.listdir("/sys/class/power_supply"):
                if not bat.startswith("BAT"):
                    continue

                with open(f"/sys/class/power_supply/{bat}/type") as f:
                    if "Battery" not in f.read():
                        continue

                self.bat = f"/sys/class/power_supply/{bat}"

        except Exception as e:
            logger.error(f"Failed to read thermal zones: {e}")

        if self.therm:
            logger.info(f"Found thermal zones:")
            for path, temp in self.therm.items():
                logger.info(f"  {path}: hot @ {temp // 1000}C")
        if self.bat:
            logger.info(f"Found battery:\n{self.bat}")

    def settings(self):
        set = {"gamemode": {"power": load_relative_yaml("power.yml")}}

        if self.win_bootnum is None:
            del set["gamemode"]["power"]["children"]["reboot_windows"]

        return set

    def update(self, conf: Config):
        if self.win_bootnum is not None and conf.get_action(
            "gamemode.power.reboot_windows"
        ):
            boot_windows()

        if conf.get_action("gamemode.power.hibernate"):
            status = emergency_hibernate(shutdown=False)
            conf["gamemode.power.status"] = status

        self.check_thermal = conf.get("gamemode.power.hibernate_auto", False)
        if self.check_thermal:
            if not self.alarm_set:
                self.alarm_set = True
                if self.bat:
                    try:
                        set_bat_alarm(self.bat)
                    except Exception as e:
                        logger.error(f"Failed to set battery alarms:\n{e}")

            curr = time.time()
            if (
                curr - self.last_check > TEMP_CHECK_INTERVAL
                and curr - self.init > TEMP_CHECK_INITIALIZE
            ):
                self.last_check = curr
                try:
                    if thermal_check(self.therm, self.bat, self.last_attempt):
                        self.last_attempt = time.time()
                except Exception as e:
                    logger.error(f"Failed to check thermal zones:\n{e}")
                    self.therm = {}
                    self.bat = None

    def notify(self, events: Sequence):
        for ev in events:
            if ev["type"] == "special" and ev.get("event", None) == "wakeup":
                delete_temporary_swap()

                if self.check_thermal:
                    try:
                        # Battery alarms neeed to be reset after hibernation
                        set_bat_alarm(self.bat)
                    except Exception as e:
                        logger.error(f"Failed to reset battery alarms:\n{e}")
                        self.bat = None
                    try:
                        if thermal_check(self.therm, self.bat, self.last_attempt, wakeup=True):
                            self.last_attempt = time.time()
                    except Exception as e:
                        logger.error(f"Failed to check thermal zones:\n{e}")
                        self.therm = {}
                        self.bat = None
                return


def autodetect(existing: Sequence[HHDPlugin]) -> Sequence[HHDPlugin]:
    if len(existing):
        return existing

    return [PowerPlugin()]



import logging
import re
import select
import time
from threading import Event as TEvent
from typing import Sequence

from hhd.controller import Button, Consumer, Event, Producer, DEBUG_MODE
from hhd.controller.lib.hide import unhide_all
from hhd.controller.base import Multiplexer, TouchpadAction
from hhd.controller.physical.evdev import B as EC
from hhd.controller.physical.evdev import GenericGamepadEvdev, enumerate_evs
from hhd.controller.virtual.uinput import HHD_PID_VENDOR, UInputDevice
from hhd.plugins import Config, Context, Emitter, get_outputs

from .const import (
    LGO_RAW_INTERFACE_AXIS_MAP,
    LGO_RAW_INTERFACE_BTN_ESSENTIALS,
    LGO_RAW_INTERFACE_BTN_MAP,
    LGO_RAW_INTERFACE_CONFIG_MAP,
    LGO_TOUCHPAD_AXIS_MAP,
    LGO_TOUCHPAD_BUTTON_MAP,
)
from .hid import LegionHidraw, RgbCallback

FIND_DELAY = 0.1
ERROR_DELAY = 0.5
LONGER_ERROR_DELAY = 3
LONGER_ERROR_MARGIN = 1.3
SELECT_TIMEOUT = 1

logger = logging.getLogger(__name__)

LEN_VID = 0x17EF
LEN_PIDS = {
    0x6182: "xinput",
    0x6183: "dinput",
    0x6184: "dual_dinput",
    0x6185: "fps",
}


def plugin_run(
    conf: Config,
    emit: Emitter,
    context: Context,
    should_exit: TEvent,
    updated: TEvent,
    dconf: dict,
    others: dict,
):
    reset = others.get("reset", False)
    init = time.perf_counter()
    repeated_fail = False

    while not should_exit.is_set():
        try:
            controller_mode = None
            pid = None
            first = True
            while not controller_mode and not should_exit.is_set():
                devs = enumerate_evs(vid=LEN_VID)
                if not devs:
                    if first:
                        first = False
                        logger.warning(f"Legion go controllers not found, waiting...")
                    time.sleep(FIND_DELAY)
                    continue

                for d in devs.values():
                    if d.get("product", None) in LEN_PIDS:
                        pid = d["product"]
                        controller_mode = LEN_PIDS[pid]
                        break
                else:
                    logger.error(
                        f"Legion go controllers not found, waiting {ERROR_DELAY}s."
                    )
                    time.sleep(ERROR_DELAY)
                    continue

            if not controller_mode:
                # If should_exit was set controller_mode will be null
                continue

            conf_copy = conf.copy()
            updated.clear()
            if (
                controller_mode == "xinput"
                and conf["xinput.mode"].to(str) != "disabled"
            ):
                logger.info("Launching emulated controller.")
                init = time.perf_counter()
                controller_loop_xinput(conf_copy, should_exit, updated, emit, reset)
            else:
                if controller_mode != "xinput":
                    logger.info(
                        f"Controllers in non-supported (yet) mode: {controller_mode}."
                    )
                else:
                    logger.info(
                        f"Controllers in xinput mode but emulation is disabled."
                    )
                init = time.perf_counter()
                controller_loop_rest(
                    controller_mode,
                    pid if pid else 2,
                    conf_copy,
                    should_exit,
                    updated,
                    emit,
                    reset,
                )
            repeated_fail = False
        except Exception as e:
            failed_fast = init + LONGER_ERROR_MARGIN > time.perf_counter()
            sleep_time = (
                LONGER_ERROR_DELAY if repeated_fail and failed_fast else ERROR_DELAY
            )
            repeated_fail = failed_fast
            logger.error(f"Received the following error:\n{type(e)}: {e}")
            logger.error(
                f"Assuming controllers disconnected, restarting after {sleep_time}s."
            )
            # Raise exception
            if DEBUG_MODE:
                raise e
            time.sleep(sleep_time)
        reset = False

    # Unhide all devices before exiting
    unhide_all()

def controller_loop_rest(
    mode: str,
    pid: int,
    conf: Config,
    should_exit: TEvent,
    updated: TEvent,
    emit: Emitter,
    reset: bool,
):
    debug = DEBUG_MODE
    shortcuts_enabled = conf["shortcuts"].to(bool)
    # FIXME: Sleep when shortcuts are disabled instead of polling raw interface
    if shortcuts_enabled:
        logger.info(f"Launching a shortcuts device.")
    else:
        logger.info(f"Shortcuts disabled. Waiting for controllers to change modes.")

    d_raw = SelectivePassthrough(
        LegionHidraw(
            vid=[LEN_VID],
            pid=list(LEN_PIDS),
            usage_page=[0xFFA0],
            usage=[0x0001],
            report_size=64,
            axis_map=LGO_RAW_INTERFACE_AXIS_MAP,
            btn_map=LGO_RAW_INTERFACE_BTN_MAP,
            required=True,
        ).with_settings(
            gyro=None, reset=reset, swap_legion=conf["swap_legion_v2"].to(bool)
        ),
        passthrough_pressed=True,
    )

    multiplexer = Multiplexer(
        dpad="both",
        trigger="analog_to_discrete",
        share_to_qam=True,
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
    )
    d_uinput = UInputDevice(
        name=f"HHD Shortcuts (Legion Mode: {mode})",
        pid=HHD_PID_VENDOR | 0x0200 | (pid & 0xF),
        phys=f"phys-hhd-shortcuts-legion-{mode}",
    )

    d_shortcuts = GenericGamepadEvdev(
        vid=[LEN_VID],
        pid=list(LEN_PIDS),
        name=[re.compile(r"Legion-Controller \d-.. Keyboard")],
        capabilities={EC("EV_KEY"): [EC("KEY_1")]},
        required=True,
    )

    try:
        fds = []
        fds.extend(d_raw.open())
        if shortcuts_enabled:
            fds.extend(d_shortcuts.open())
            fds.extend(d_uinput.open())

        while not should_exit.is_set() and not updated.is_set():
            select.select(fds, [], [], SELECT_TIMEOUT)
            evs = multiplexer.process(d_raw.produce(fds))

            if shortcuts_enabled:
                d_shortcuts.produce(fds)
                d_uinput.produce(fds)
                if debug and evs:
                    logger.info(evs)
                d_uinput.consume(evs)
    finally:
        d_uinput.close(True)
        d_shortcuts.close(True)
        d_raw.close(True)


def controller_loop_xinput(
    conf: Config, should_exit: TEvent, updated: TEvent, emit: Emitter, reset: bool
):
    debug = DEBUG_MODE

    # Output
    dimu = conf["imu.mode"].to(str)

    match dimu:
        case "left":
            simu = "left_to_main"
            cidx = 1
        case "right" | "both":
            simu = "right_to_main"
            cidx = 2
        case _:
            simu = None
            cidx = 0

    d_producers, d_outs, d_params = get_outputs(
        conf["xinput"],
        conf["touchpad"],
        dimu != "disabled",
        controller_id=cidx,
        emit=emit,
        dual_motion=dimu == "both",
        rgb_modes={
            "disabled": [],
            "solid": ["color"],
            "pulse": ["color", "speed"],
            "rainbow": ["brightness", "speed"],
            "spiral": ["brightness", "speed"],
        },
    )
    motion = d_params.get("uses_motion", True)
    dual_motion = d_params.get("uses_dual_motion", True)
    swap_legion = conf["swap_legion_v2"].to(bool)
    if not dual_motion and dimu == "both":
        dimu = "right"
    if not motion:
        dimu = "disabled"

    # Inputs
    d_xinput = GenericGamepadEvdev(
        vid=[0x17EF],
        pid=[0x6182],
        # name=["Generic X-Box pad"],
        capabilities={EC("EV_KEY"): [EC("BTN_A")]},
        required=True,
        hide=True,
    )
    d_touch = GenericGamepadEvdev(
        vid=[0x17EF],
        pid=[0x6182],
        name=[re.compile(".+Touchpad")],  # "  Legion Controller for Windows  Touchpad"
        capabilities={EC("EV_KEY"): [EC("BTN_MOUSE")]},
        btn_map=LGO_TOUCHPAD_BUTTON_MAP,
        axis_map=LGO_TOUCHPAD_AXIS_MAP,
        aspect_ratio=1,
        required=True,
    )
    d_raw = SelectivePassthrough(
        LegionHidraw(
            vid=[LEN_VID],
            pid=list(LEN_PIDS),
            usage_page=[0xFFA0],
            usage=[0x0001],
            report_size=64,
            axis_map=LGO_RAW_INTERFACE_AXIS_MAP,
            btn_map=LGO_RAW_INTERFACE_BTN_MAP,
            config_map=LGO_RAW_INTERFACE_CONFIG_MAP,
            callback=RgbCallback(),
            required=True,
        ).with_settings(
            gyro=dimu,
            reset=reset,
            use_touchpad=False,
            swap_legion=swap_legion,
        )
    )

    # Mute keyboard shortcuts, mute
    d_shortcuts = GenericGamepadEvdev(
        vid=[LEN_VID],
        pid=list(LEN_PIDS),
        name=[re.compile(".+Keyboard")],  # "  Legion Controller for Windows  Keyboard"
        # capabilities={EC("EV_KEY"): [EC("KEY_1")]},
        # report_size=64,
        required=True,
    )

    touch_actions = (
        conf["touchpad.controller"]
        if conf["touchpad.mode"].to(TouchpadAction) == "controller"
        else conf["touchpad.emulation"]
    )
    multiplexer = Multiplexer(
        trigger="analog_to_discrete",
        dpad="both",
        led="main_to_sides",
        status="both_to_main",
        share_to_qam=True,
        swap_guide="guide_is_select" if swap_legion else None,
        touchpad_short=touch_actions["short"].to(TouchpadAction),
        touchpad_right=touch_actions["hold"].to(TouchpadAction),
        select_reboots=conf["select_reboots"].to(bool),
        r3_to_share=conf["m2_to_mute"].to(bool),
        nintendo_mode=conf["nintendo_mode"].to(bool),
        emit=emit,
        imu=simu,
        params=d_params,
    )

    REPORT_FREQ_MIN = 25
    REPORT_FREQ_MAX = 500

    REPORT_DELAY_MAX = 1 / REPORT_FREQ_MIN
    REPORT_DELAY_MIN = 1 / REPORT_FREQ_MAX

    fds = []
    devs = []
    fd_to_dev = {}

    def prepare(m):
        devs.append(m)
        fs = m.open()
        fds.extend(fs)
        for f in fs:
            fd_to_dev[f] = m

    try:
        prepare(d_xinput)
        prepare(d_shortcuts)
        if d_params["uses_touch"]:
            prepare(d_touch)
        prepare(d_raw)
        for d in d_producers:
            prepare(d)

        ts_count: dict[str, int] = {"left_imu_ts": 0, "right_imu_ts": 0}
        ts_last: dict[str, int] = {"left_imu_ts": 0, "right_imu_ts": 0}

        logger.info("Emulated controller launched, have fun!")
        while not should_exit.is_set() and not updated.is_set():
            start = time.perf_counter()
            # Add timeout to call consumers a minimum amount of times per second
            r, _, _ = select.select(fds, [], [], REPORT_DELAY_MAX)
            evs = []
            to_run = set()
            for f in r:
                to_run.add(id(fd_to_dev[f]))

            for d in devs:
                if id(d) in to_run:
                    evs.extend(d.produce(r))

            # Patch timestamps to convert them to ns
            # for d in ('x', 'y', 'z'):
            #     p = False
            #     for ev in evs:
            #         if f"accel_{d}" in ev["code"]:
            #             print(
            #                 f"{ev['code'].split('accel_')[1]}: {ev['value']:12.5e} ", end=""
            #             )
            #             p = True
            #     if not p:
            #         print(f"{d}:              ", end='')
            # print()
            for ev in evs:
                if ev["type"] == "axis" and "_imu_ts" in ev["code"]:
                    # Find diff between previous event
                    last = ts_last[ev["code"]]
                    curr = ev["value"]
                    diff = curr - last
                    if curr < last:
                        diff += 256
                    ts_last[ev["code"]] = curr
                    # 8ms per count
                    ts_count[ev["code"]] += diff * 8_000_000
                    ev["value"] = ts_count[ev["code"]]
                if ev["type"] == "axis" and "gyro" in ev["code"]:
                    v = ev["value"]
                    if (abs(v / 0.001065) // 1) in (254, 255):
                        # Legion go controllers have a bug where they will
                        # randomly output 254 or 255. If that happens, drop event
                        ev["code"] = ""  # type: ignore

            evs = multiplexer.process(evs)
            if evs:
                if debug:
                    logger.info(evs)

                d_xinput.consume(evs)
                d_raw.consume(evs)

            for d in d_outs:
                d.consume(evs)

            t = time.perf_counter()
            elapsed = t - start
            if elapsed < REPORT_DELAY_MIN:
                time.sleep(REPORT_DELAY_MIN - elapsed)

    except KeyboardInterrupt:
        raise
    finally:
        for d in reversed(devs):
            try:
                d.close(not updated.is_set())
            except Exception as e:
                logger.error(f"Error while closing device '{d}' with exception:\n{e}")
                if debug:
                    raise e


class SelectivePassthrough(Producer, Consumer):

    def __init__(
        self,
        parent,
        forward_buttons: Sequence[Button] = ("share", "mode"),
        passthrough: Sequence[Button] = list(
            next(iter(LGO_RAW_INTERFACE_BTN_ESSENTIALS.values()))
        ),
        passthrough_pressed: bool = False,
    ):
        self.parent = parent
        self.state = False

        self.forward_buttons = forward_buttons
        self.passthrough = passthrough
        self.pressed_time = None
        self.pressed_vals = set()
        self.passthrough_pressed = passthrough_pressed

        self.to_disable_btn = set()
        self.to_disable_axis = set()

    def open(self) -> Sequence[int]:
        return self.parent.open()

    def close(self, exit: bool) -> bool:
        return self.parent.close(exit)

    def produce(self, fds: Sequence[int]) -> Sequence[Event]:
        evs: Sequence[Event] = self.parent.produce(fds)

        out = []
        curr = time.perf_counter()
        if self.passthrough_pressed:
            passthrough = bool(self.pressed_vals)
        else:
            passthrough = self.pressed_time and (curr - self.pressed_time < 1)

        for ev in evs:
            if ev["type"] == "button" and ev["code"] in self.forward_buttons:
                if ev.get("value", False):
                    self.pressed_time = curr
                    self.pressed_vals.add(ev["code"])
                else:
                    self.pressed_vals.discard(ev["code"])

            if ev["type"] == "configuration":
                out.append(ev)
            elif ev["type"] == "button" and ev["code"] in self.passthrough:
                out.append(ev)
            elif ev["type"] == "axis" and (
                "imu" in ev["code"] or "accel" in ev["code"] or "gyro" in ev["code"]
            ):
                out.append(ev)
            elif "touchpad" in ev["code"]:
                out.append(ev)

        if passthrough:
            # If mode is pressed, forward all events
            return evs
        else:
            return out

    def consume(self, events: Sequence[Event]):
        return self.parent.consume(events)



import argparse
import fcntl
import logging
import os
import signal
import subprocess
import sys
import time
from os.path import join
from threading import Condition
from threading import Event as TEvent
from threading import RLock
from time import sleep
from typing import Sequence

import pkg_resources

from .logging import set_log_plugin, setup_logger, update_log_plugins
from .plugins import (
    Config,
    Emitter,
    Event,
    HHDAutodetect,
    HHDPlugin,
    HHDSettings,
    load_relative_yaml,
)
from .plugins.settings import (
    Validator,
    get_default_state,
    get_settings_hash,
    load_blacklist_yaml,
    load_profile_yaml,
    load_state_yaml,
    merge_settings,
    parse_defaults,
    save_blacklist_yaml,
    save_profile_yaml,
    save_state_yaml,
    validate_config,
)
from .utils import (
    GIT_ADJ,
    GIT_HHD,
    HHD_DEV_DIR,
    expanduser,
    fix_perms,
    get_ac_status,
    get_ac_status_fn,
    get_context,
    get_os,
    switch_priviledge,
)

logger = logging.getLogger(__name__)

CONFIG_DIR = os.environ.get("HHD_CONFIG_DIR", "~/.config/hhd")
PLUGIN_WHITELIST = os.environ.get("HHD_PLUGINS", "")

ERROR_DELAY = 5
INIT_DELAY = 0.4
POLL_DELAY = 2
SLEEP_MIN_T = 8


class EmitHolder(Emitter):
    def __init__(self, condition: Condition, ctx, info) -> None:
        self._events = []
        self._condition = condition
        super().__init__(ctx=ctx, info=info)

    def __call__(self, event: Event | Sequence[Event]) -> None:
        with self._condition:
            if isinstance(event, Sequence):
                self._events.extend(event)
            else:
                self._events.append(event)
            self._condition.notify_all()

    def get_events(self, timeout: int = -1) -> Sequence[Event]:
        with self._condition:
            if not self._events and timeout != -1:
                self._condition.wait()
            ev = self._events
            self._events = []
            return ev

    def has_events(self):
        with self._condition:
            return bool(self._events)


def notifier(ev: TEvent, cond: Condition):
    def _inner(sig, frame):
        with cond:
            ev.set()
            cond.notify_all()

    return _inner


def get_wakeup_count():
    try:
        with open("/sys/power/wakeup_count", "r") as f:
            return int(f.read().strip())
    except Exception:
        return -1


def print_token(ctx):
    token_fn = expanduser(join(CONFIG_DIR, "token"), ctx)

    try:
        with open(token_fn, "r") as f:
            token = f.read().strip()

        logger.info(f'Current HHD token (for user "{ctx.name}") is: "{token}"')
    except Exception as e:
        logger.error(f"Token not found or could not be read, error:\n{e}")
        logger.info(
            "Enable the http endpoint to generate a token automatically.\n"
            + "Or place it under '~/.config/hhd/token' manually.\n"
            + "'chown 600 ~/.config/hhd/token' for security reasons!"
        )


def main():
    parser = argparse.ArgumentParser(
        prog="HHD: Handheld Daemon main interface.",
        description="Handheld Daemon is a daemon for managing the quirks inherent in handheld devices.",
    )
    parser.add_argument(
        "-u",
        "--user",
        default=None,
        help="The user whose home directory will be used to store the files (~/.config/hhd).",
        dest="user",
    )
    parser.add_argument(
        "command",
        nargs="*",
        default=[],
        help="The command to run. If empty, run as daemon. Right now, only the command `token` is supported.",
    )
    args = parser.parse_args()
    user = args.user

    # Setup temporary logger for permission retrieval
    ctx = get_context(user)
    if not ctx:
        print(f"Could not get user information. Exiting...")
        return

    detectors: dict[str, HHDAutodetect] = {}
    plugins: dict[str, Sequence[HHDPlugin]] = {}
    cfg_fds = []
    switch_root = None

    # HTTP data
    https = None
    prev_http_cfg = None
    updated = False
    last_event = None
    ac_fn = get_ac_status_fn()
    info = Config()
    info["ac"] = None
    ac_status = None

    # Check we are in a virtual environment
    # TODO: Improve
    exe_python = sys.executable

    try:
        # Create nested hhd dir
        # This might mess up permissions in upward directories
        # So try to deescalate
        hhd_dir = expanduser(CONFIG_DIR, ctx)
        try:
            switch_priviledge(ctx, False)
            os.makedirs(hhd_dir, exist_ok=True)
            switch_priviledge(ctx, True)
            fix_perms(hhd_dir, ctx)
        except Exception:
            pass

        # Remove old dir
        try:
            os.rename(
                join(hhd_dir, "plugins"), join(hhd_dir, "plugins_old_USE_STATEYML")
            )
        except Exception:
            pass

        set_log_plugin("main")
        setup_logger(join(CONFIG_DIR, "log"), ctx=ctx)

        if args.command:
            if args.command[0] == "token":
                print_token(ctx)
                return
            else:
                logger.error(f"Command '{args.command[0]}' is unknown. Ignoring...")

        # Get OS Info
        info["os"] = get_os()

        # Use blacklist
        blacklist_fn = join(hhd_dir, "plugins.yml")
        blacklist = load_blacklist_yaml(blacklist_fn)

        logger.info(f"Running autodetection...")

        detector_names = []
        whitelist = PLUGIN_WHITELIST.split(",") if PLUGIN_WHITELIST else []
        for autodetect in pkg_resources.iter_entry_points("hhd.plugins"):
            name = autodetect.name
            detector_names.append(name)
            if name in blacklist:
                logger.info(f"Skipping blacklisted provider '{name}'.")
            if whitelist and name not in whitelist:
                logger.info(f"Skipping provider '{name}' due to whitelist.")
                continue

            detectors[autodetect.name] = autodetect.resolve()

        # Save new blacklist file
        save_blacklist_yaml(blacklist_fn, detector_names, blacklist)
        fix_perms(blacklist_fn, ctx)

        logger.info(f"Found plugin providers: {', '.join(list(detectors))}")

        for name, autodetect in detectors.items():
            plugins[name] = autodetect([])

        plugin_str = "Loaded the following plugins:"
        for pkg_name, sub_plugins in plugins.items():
            if not sub_plugins:
                continue
            plugin_str += (
                f"\n  - {pkg_name:>8s}: {', '.join(p.name for p in sub_plugins)}"
            )
        logger.info(plugin_str)

        # Get sorted plugins
        sorted_plugins: Sequence[HHDPlugin] = []
        for plugs in plugins.values():
            sorted_plugins.extend(plugs)
        sorted_plugins.sort(key=lambda x: x.priority)
        validator: Validator = lambda tags, config, value: any(
            p.validate(tags, config, value) for p in sorted_plugins
        )

        if not sorted_plugins:
            logger.error(f"No plugins started, exiting...")
            return

        # Load locales
        locales = []
        for register in pkg_resources.iter_entry_points("hhd.i18n"):
            locales.extend(register.resolve()())
        locales.sort(key=lambda x: x["priority"], reverse=True)

        if locales:
            lstr = "Loaded the following locales:\n"
            for locale in locales:
                lstr += (
                    f" - {locale['domain']} ({locale['priority']}): {locale['dir']}\n"
                )
            logger.info(lstr[:-1])
        else:
            logger.info("No locales found.")

        # Open plugins
        lock = RLock()
        cond = Condition(lock)
        emit = EmitHolder(cond, ctx, info)
        for p in sorted_plugins:
            set_log_plugin(getattr(p, "log") if hasattr(p, "log") else "ukwn")
            p.open(emit, ctx)
            update_log_plugins()
        set_log_plugin("main")

        # Compile initial configuration
        state_fn = expanduser(join(CONFIG_DIR, "state.yml"), ctx)
        token_fn = expanduser(join(CONFIG_DIR, "token"), ctx)
        settings: HHDSettings = {}
        shash = None

        # Load profiles
        profiles = {}
        templates = {}
        conf = Config({})
        profile_dir = expanduser(join(CONFIG_DIR, "profiles"), ctx)
        os.makedirs(profile_dir, exist_ok=True)
        fix_perms(profile_dir, ctx)

        # Monitor config files for changes
        should_initialize = TEvent()
        initial_run = True
        reset = False
        should_exit = TEvent()
        signal.signal(signal.SIGPOLL, notifier(should_initialize, cond))
        signal.signal(signal.SIGINT, notifier(should_exit, cond))
        signal.signal(signal.SIGTERM, notifier(should_exit, cond))

        # Get wakeup count for sleep detection
        wakeup_count = get_wakeup_count()

        while not should_exit.is_set():
            #
            # Configuration
            #

            # Initialize if files changed
            if should_initialize.is_set() or initial_run:
                # wait a bit to allow other processes to save files
                if not initial_run:
                    sleep(INIT_DELAY)
                initial_run = False
                set_log_plugin("main")
                logger.info(f"Reloading configuration.")

                # Settings
                hhd_settings = {"hhd": load_relative_yaml("settings.yml")}
                # TODO: Improve check
                try:
                    if "venv" not in exe_python:
                        del hhd_settings["hhd"]["settings"]["children"]["update_stable"]
                        del hhd_settings["hhd"]["settings"]["children"]["update_beta"]
                except Exception as e:
                    logger.warning(f"Could not hide update settings. Error:\n{e}")
                settings = merge_settings(
                    [hhd_settings, *[p.settings() for p in sorted_plugins]]
                )
                # Force general settings to be last
                if "hhd" in settings:
                    settings = dict(settings)
                    tmp = settings.pop("hhd")
                    settings["hhd"] = tmp
                shash = get_settings_hash(settings)

                # State
                if reset:
                    logger.warning(f"Resetting settings.")
                    conf = get_default_state(settings)
                    conf.updated = True
                    reset = False
                else:
                    new_conf = load_state_yaml(state_fn, settings)
                    if not new_conf:
                        if conf.conf:
                            logger.warning(f"Using previous configuration.")
                        else:
                            logger.info(f"Using default configuration.")
                            conf = get_default_state(settings)
                    else:
                        conf = new_conf

                from importlib.metadata import version

                try:
                    ver = version("hhd")
                    conf["hhd.settings.version"] = ver
                    logger.info(f"Handheld Daemon Version: {ver}")
                except Exception:
                    pass

                try:
                    ver = version("adjustor")
                    conf["hhd.settings.version_adj"] = ver
                    logger.info(f"Adjustor Version: {ver}")
                except Exception:
                    conf["hhd.settings.version_adj"] = "Not Installed"
                    logger.info(f"Adjustor not installed")

                try:
                    from hhd.plugins.overlay.overlay import (
                        find_overlay_exe,
                        get_overlay_version,
                    )

                    exe = find_overlay_exe(ctx)
                    if exe:
                        ver = get_overlay_version(exe, ctx)
                        conf["hhd.settings.version_ui"] = ver
                        logger.info(f"Overlay Version: {ver}")
                    else:
                        conf["hhd.settings.version_ui"] = "Not Installed"
                        logger.info(f"Overlay not installed")
                except Exception:
                    logger.info(exe)

                # Profiles
                profiles = {}
                templates = {}
                os.makedirs(profile_dir, exist_ok=True)
                fix_perms(profile_dir, ctx)
                for fn in os.listdir(profile_dir):
                    if not fn.endswith(".yml"):
                        continue
                    name = fn.replace(".yml", "")
                    s = load_profile_yaml(join(profile_dir, fn))
                    if s:
                        validate_config(s, settings, validator, use_defaults=False)
                        if name.startswith("_"):
                            templates[name] = s
                        else:
                            # Profiles are shared so lock when accessing
                            # Configs have their own locks and are safe
                            with lock:
                                profiles[name] = s
                if profiles:
                    logger.info(
                        f"Loaded the following profiles (and state):\n[{', '.join(profiles)}]"
                    )
                else:
                    logger.info(f"No profiles found.")

                # Monitor files for changes
                for fd in cfg_fds:
                    try:
                        fcntl.fcntl(fd, fcntl.F_NOTIFY, 0)
                        os.close(fd)
                    except Exception:
                        pass
                cfg_fds = []
                cfg_fns = [
                    CONFIG_DIR,
                    join(CONFIG_DIR, "profiles"),
                ]
                for fn in cfg_fns:
                    fd = -1
                    try:
                        fd = os.open(expanduser(fn, ctx), os.O_RDONLY)
                        fcntl.fcntl(
                            fd,
                            fcntl.F_NOTIFY,
                            fcntl.DN_CREATE
                            | fcntl.DN_DELETE
                            | fcntl.DN_MODIFY
                            | fcntl.DN_RENAME
                            | fcntl.DN_MULTISHOT,
                        )
                    except Exception:
                        if fd != -1:
                            os.close(fd)
                        continue
                    cfg_fds.append(fd)

                should_initialize.clear()
                logger.info(f"Initialization Complete!")

            # Initialize http server
            http_cfg = conf["hhd.http"]
            if http_cfg != prev_http_cfg:
                prev_http_cfg = http_cfg
                if https:
                    https.close()
                if http_cfg["enable"].to(bool):
                    from .http import HHDHTTPServer

                    port = http_cfg["port"].to(int)
                    localhost = http_cfg["localhost"].to(bool)
                    use_token = http_cfg["token"].to(bool)

                    # Generate security token
                    if use_token:
                        if not os.path.isfile(token_fn):
                            import hashlib
                            import random

                            token = hashlib.sha256(
                                str(random.random()).encode()
                            ).hexdigest()[:12]
                            with open(token_fn, "w") as f:
                                os.chmod(token_fn, 0o600)
                                f.write(token)
                            fix_perms(token_fn, ctx)
                        else:
                            with open(token_fn, "r") as f:
                                token = f.read().strip()
                    else:
                        token = None

                    set_log_plugin("rest")
                    https = HHDHTTPServer(localhost, port, token)
                    https.update(settings, conf, info, profiles, emit, locales, ctx)
                    try:
                        https.open()
                    except Exception as e:
                        logger.error(
                            f"Could not start http API on port {port}.\n"
                            + "Is another version of Handheld Daemon open?\n"
                            + "Closing."
                        )
                        return
                    update_log_plugins()
                    set_log_plugin("main")

            #
            # Plugin loop
            #

            # Process events
            set_log_plugin("main")
            settings_changed = False
            events = emit.get_events()

            new_wakeup_count = get_wakeup_count()
            curr = time.time()
            # Debounce sleep event to avoid spurious wakeup triggers
            # This loop will run every 2 seconds, perhaps 4 seconds if there is
            # a delay. Unless 8 seconds lapse, ignore the event
            if (
                new_wakeup_count != wakeup_count
                and last_event
                and curr > last_event + SLEEP_MIN_T
            ):
                logger.info(
                    f"System woke up from sleep. Wakeup count: {new_wakeup_count} from {wakeup_count}."
                )
                events: Sequence[Event] = [
                    *events,
                    {
                        "type": "special",
                        "event": "wakeup",
                        "data": {
                            "count": new_wakeup_count
                        },  # FIXME: Count might be removed in the future
                    },
                ]
            wakeup_count = new_wakeup_count
            last_event = curr

            # AC status
            if ac_fn:
                new_status = get_ac_status(ac_fn)
                if new_status != ac_status:
                    logger.info(f"AC status is: {new_status}")
                    ac_status = new_status
                    info["ac"] = ac_status

            for ev in events:
                match ev["type"]:
                    case "settings":
                        settings_changed = True
                    case "profile":
                        new_conf = ev["config"]
                        if new_conf:
                            with lock:
                                profiles[ev["name"]] = ev["config"]
                            validate_config(
                                profiles[ev["name"]],
                                settings,
                                validator,
                                use_defaults=False,
                            )
                        else:
                            with lock:
                                if ev["name"] in profiles:
                                    del profiles[ev["name"]]
                    case "apply":
                        if ev["name"] in profiles:
                            conf.update(profiles[ev["name"]].conf)
                    case "state":
                        conf.update(ev["config"].conf)
                    case "special":
                        if ev["event"] == "restart_dev":
                            should_exit.set()
                            switch_root = True
                            break
                        elif ev["event"] == "shutdown_dev":
                            should_exit.set()
                            # Trigger restart
                            updated = True
                    case "acpi" | "tdp" | "ppd" | "energy":
                        pass
                    case other:
                        logger.error(f"Invalid event type submitted: '{other}'")

            # If settings changed, the configuration needs to reload
            # but it needs to be saved first
            if settings_changed:
                logger.info(f"Reloading settings.")

                # Settings
                settings_base = {k: {} for k in load_relative_yaml("sections.yml")['sections']}
                hhd_settings = {"hhd": load_relative_yaml("settings.yml")}
                # TODO: Improve check
                try:
                    if "venv" not in exe_python:
                        del hhd_settings["hhd"]["settings"]["children"]["update_stable"]
                        del hhd_settings["hhd"]["settings"]["children"]["update_beta"]
                except Exception as e:
                    logger.warning(f"Could not hide update settings. Error:\n{e}")
                settings = merge_settings(
                    [
                        settings_base,
                        hhd_settings,
                        *[p.settings() for p in sorted_plugins],
                    ]
                )
                # Force general settings to be last
                if "hhd" in settings:
                    settings = dict(settings)
                    tmp = settings.pop("hhd")
                    settings["hhd"] = tmp
                shash = get_settings_hash(settings)

                # Add new defaults
                conf = Config([parse_defaults(settings), conf.conf])
                conf.updated = True

            # Validate config
            validate_config(conf, settings, validator)

            #
            # Plugin event loop
            #

            # Allow plugins to process events
            if events:
                for p in sorted_plugins:
                    set_log_plugin(getattr(p, "log") if hasattr(p, "log") else "ukwn")
                    p.notify(events)
                    update_log_plugins()

            # Run prepare loop
            for p in reversed(sorted_plugins):
                set_log_plugin(getattr(p, "log") if hasattr(p, "log") else "ukwn")
                p.prepare(conf)
                update_log_plugins()

            # Run update loop
            for p in sorted_plugins:
                set_log_plugin(getattr(p, "log") if hasattr(p, "log") else "ukwn")
                p.update(conf)
                update_log_plugins()
            set_log_plugin("ukwn")

            # Notify that events were applied
            # Before saving to reduce delay (yaml files take 100ms :( )
            if https:
                https.update(settings, conf, info, profiles, emit, locales, ctx)

            #
            # Save loop
            #

            has_new = should_initialize.is_set()
            saved = False
            # Save existing profiles if open
            if save_state_yaml(state_fn, settings, conf, shash):
                fix_perms(state_fn, ctx)
                saved = True
                conf.updated = False
            for name, prof in profiles.items():
                fn = join(profile_dir, name + ".yml")
                if save_profile_yaml(fn, settings, prof, shash):
                    fix_perms(fn, ctx)
                    saved = True
                    prof.updated = False
            for prof in os.listdir(profile_dir):
                if prof.startswith("_") or not prof.endswith(".yml"):
                    continue
                name = prof[:-4]
                if name not in profiles:
                    fn = join(profile_dir, prof)
                    try:
                        new_fn = fn + ".bak"
                        os.rename(fn, new_fn)
                        saved = True
                    except Exception as e:
                        logger.error(
                            f"Failed removing profile {name} at:\n{fn}\nWith error:\n{e}"
                        )

            # Causes unnecessary writes, is not used anyway.
            # # Add template config
            # if save_profile_yaml(
            #     join(profile_dir, "_template.yml"),
            #     settings,
            #     templates.get("_template", None),
            #     shash,
            # ):
            #     fix_perms(join(profile_dir, "_template.yml"), ctx)
            #     saved = True

            if not has_new and saved:
                # We triggered the interrupt, clear
                should_initialize.clear()

            upd_stable = conf.get("hhd.settings.update_stable", False)
            upd_beta = conf.get("hhd.settings.update_beta", False)

            if upd_stable or upd_beta:
                set_log_plugin("main")
                conf["hhd.settings.update_stable"] = False
                conf["hhd.settings.update_beta"] = False

                switch_priviledge(ctx, False)
                try:
                    logger.info(f"Updating Handheld Daemon.")
                    if "venv" in exe_python:
                        subprocess.check_call(
                            [
                                exe_python,
                                "-m",
                                "pip",
                                "uninstall",
                                "-y",
                                "hhd",
                                "adjustor",
                            ]
                        )
                        subprocess.check_call(
                            [
                                exe_python,
                                "-m",
                                "pip",
                                "install",
                                "--upgrade",
                                "--cache-dir",
                                "/tmp/__hhd_update_cache",
                                (GIT_HHD if upd_beta else "hhd"),
                                (GIT_ADJ if upd_beta else "adjustor"),
                            ]
                        )

                        if not upd_beta:
                            # No beta version for the UI yet, skip updating it
                            import json
                            import stat
                            import urllib.request

                            with urllib.request.urlopen(
                                "https://api.github.com/repos/hhd-dev/hhd-ui/releases/latest"
                            ) as f:
                                release_data = json.load(f)

                            for asset in release_data["assets"]:
                                os.makedirs(
                                    expanduser("~/.local/bin", ctx), exist_ok=True
                                )
                                if "hhd-ui.AppImage" == asset["name"]:
                                    REPORT_POINTS = 8

                                    def progress(idx, blockSize, total):
                                        interval = int(
                                            total / blockSize / REPORT_POINTS
                                        )
                                        if idx % interval == 0:
                                            logger.info(
                                                f"Downloading overlay: {100*idx*blockSize / total:.1f}%"
                                            )

                                    out_fn = expanduser("~/.local/bin/hhd-ui", ctx)
                                    urllib.request.urlretrieve(
                                        asset["browser_download_url"],
                                        out_fn,
                                        reporthook=progress,
                                    )

                                    st = os.stat(out_fn)
                                    os.chmod(out_fn, st.st_mode | stat.S_IEXEC)
                                    break

                        # Set updated
                        updated = True
                    else:
                        logger.error(
                            f"Could not update, python executable is not within a venv (checked for 'venv' in path name):\n{exe_python}"
                        )
                except Exception as e:
                    err = f"Error while updating:\n{e}"
                    conf["hhd.settings.update_error"] = err
                    logger.error(err)
                switch_priviledge(ctx, True)

                if updated:
                    should_exit.set()

            # Wait for events
            with lock:
                if (
                    not should_exit.is_set()
                    and not settings_changed
                    and not should_initialize.is_set()
                    and not emit.has_events()
                ):
                    cond.wait(timeout=POLL_DELAY)

            # Check reset
            if conf["hhd.settings.reset"].to(bool):
                conf["hhd.settings.reset"] = False
                should_initialize.set()
                reset = True

        set_log_plugin("main")
        logger.info(f"Received interrupt or updated. Stopping plugins and exiting.")
    finally:
        for fd in cfg_fds:
            try:
                os.close(fd)
            except Exception:
                pass
        if https:
            set_log_plugin("main")
            logger.info("Shutting down the REST API.")
            https.close()
        for plugs in plugins.values():
            for p in plugs:
                set_log_plugin("main")
                logger.info(f"Stopping plugin `{p.name}`.")
                set_log_plugin(getattr(p, "log") if hasattr(p, "log") else "ukwn")
                p.close()

        set_log_plugin("main")
        try:
            logger.info("Closing cached controllers.")
            from hhd.controller.virtual.dualsense import Dualsense
            from hhd.controller.virtual.uinput import UInputDevice

            UInputDevice.close_cached()
            Dualsense.close_cached()
        except Exception as e:
            logger.error("Could not close cached controllers with error:\n{e}")

    if updated:
        # Use error code to restart service
        sys.exit(-1)

    if switch_root:
        os.environ["HHD_SWITCH_ROOT"] = "1"
        o = subprocess.run([HHD_DEV_DIR + "/bin/hhd", *sys.argv], check=False)
        sys.exit(o.returncode)


if __name__ == "__main__":
    main()



from .conf import Config
from .inputs import gen_gyro_state, get_gyro_config, get_gyro_state, get_touchpad_config
from .outputs import (
    fix_limits,
    get_limits,
    get_limits_config,
    get_outputs,
    get_outputs_config,
)
from .plugin import (
    Context,
    Emitter,
    Event,
    HHDAutodetect,
    HHDLocale,
    HHDLocaleRegister,
    HHDPlugin,
)
from .settings import HHDSettings
from .utils import get_relative_fn, load_relative_yaml

__all__ = [
    "Config",
    "HHDSettings",
    "HHDAutodetect",
    "HHDPlugin",
    "get_relative_fn",
    "load_relative_yaml",
    "Emitter",
    "Event",
    "Context",
    "get_outputs_config",
    "get_outputs",
    "get_touchpad_config",
    "get_gyro_config",
    "get_gyro_state",
    "gen_gyro_state",
    "HHDLocale",
    "HHDLocaleRegister",
    "get_limits_config",
    "get_limits",
    "fix_limits",
]
